"The ActorManager is the wrapper and support structure for an actual Actor instance."

import logging
from thespian.actors import *
from thespian.system.systemBase import thesplog
from thespian.system.systemCommon import systemCommonBase
from thespian.system.addressManager import CannotPickleAddress
from thespian.system.transport import *
from thespian.system.messages.admin import *
from thespian.system.messages import *


#KWQ: target actor requirements (in PendingActor)

class ActorManager(systemCommonBase):
    def __init__(self, childClass, endpoint, transportClass, sourceHash, parentAddr, adminAddr,
                 childRequirements):
        super(ActorManager, self).__init__(adminAddr, transportClass, endpoint)
        self._parentAddr = parentAddr
        self._sourceHash = sourceHash
        # import sourceHash to get childClass
        self._actorClass = childClass
        #self._childReqs  = childRequirements
        self.actorInst   = None
        thesplog('Starting Actor %s at %s', childClass, self.transport.myAddress,
                 level = logging.INFO)



    def _createInstance(self):
        children = list(self._addrManager.sendingAddresses())
        try:
            aClass = eval(self._actorClass) if type(self._actorClass) == type("") else self._actorClass
            try:
                actorInst = aClass(childActors = None if not children else children)
            except TypeError as te:
                if "unexpected keyword argument 'childActors'" in str(te):
                    # Some actors do not take a childActors keyword argument, so init without it
                    actorInst = aClass()
                else:
                    thesplog("Got a TypeError attempting to instantiate Actor %s: %s %s",
                             self._actorClass, str(te), level=logging.WARNING)
                    return
        except Exception as ex:
            thesplog('Actor %s @ %s instantiation exception: %s', self._actorClass,
                     self.transport.myAddress, str(ex),
                     level=logging.ERROR)
            self._sayGoodbye()
            return

        self.actorInst = actorInst
        self.actorInst._myRef = self


    def run(self):
        if self.actorInst is None: self._createInstance()
        if self.actorInst:
            try:
                self.transport.run(self.handleMessages)
            except Exception as ex:
                import traceback
                thesplog('Actor %s @ %s transport exception: %s',
                         self._actorClass, self.transport.myAddress,
                         traceback.format_exc(),
                         level=logging.ERROR, exc_info=True)
        thesplog('Run %s done', self._actorClass, level=logging.DEBUG)


    def handleMessages(self, envelope):
        r = self._handleOneMessage(envelope)
        while r and self._receiveQueue:
            r = self._handleOneMessage(self._receiveQueue.pop(0))
        return r


    def _handleOneMessage(self, envelope):
        # handle admin stuff

        msg = envelope.message
        if isinstance(msg, TellMessage):
            msg = msg.actualMessage

        thesplog('ACTOR got %s from %s', msg, envelope.sender, level=logging.DEBUG)

        handled, result = self._handleReplicatorMessages(envelope)
        if handled:
            return result

        if isinstance(msg, PendingActorResponse):
            return self._pendingActorResponse(envelope)

        if not getattr(self, '_exiting', False) or \
           isinstance(msg, (ActorExitRequest, ChildActorExited)):
            try:
                self.actorInst.receiveMessage(msg, envelope.sender)
            except Exception:
                if not isinstance(msg, ActorExitRequest):
                    import traceback
                    thesplog('Actor %s @ %s retryable exception on message %s: %s',
                             self._actorClass, self.transport.myAddress, msg,
                             traceback.format_exc(),
                             level=logging.ERROR, exc_info=True)
                    try:
                        self.actorInst.receiveMessage(msg, envelope.sender)
                    except Exception:
                        import traceback
                        thesplog('Actor %s @ %s second exception on message %s: %s',
                                 self._actorClass, self.transport.myAddress, msg,
                                 traceback.format_exc(),
                                 level=logging.ERROR, exc_info=True)
                        # Try reconstituting the Actor.  This will destroy its internal history
                        # KWQ: pass existing children?  probably not... not enough information to reconstitute information, but then should the children be killed??
                        self.actorInst = None
                        self._createInstance()
                        try:
                            self.actorInst.receiveMessage(msg, envelope.sender)
                        except Exception:
                            import traceback
                            thesplog('Actor %s @ %s third exception on POISON message %s: %s',
                                     self._actorClass, self.transport.myAddress, msg,
                                     traceback.format_exc(),
                                     level=logging.ERROR, exc_info=True)
                            self.actorInst = None
                            self._createInstance()
                            if not isinstance(envelope.message, TellMessage):
                                self.transport.scheduleTransmit(
                                    TransmitIntent(envelope.sender, PoisonMessage(msg)))

        if isinstance(envelope.message, ActorExitRequest):
            if getattr(self, '_exiting', None) is not None:
                return True  # multiple shutdown requests ignored
            # Initiate exit; may be a delay while children are shutdown
            return self._actorExit(envelope.message)

        if isinstance(envelope.message, ChildActorExited):
            return self._handleChildExited(envelope.message.childAddress)

        return True


    def _actorExit(self, exitRequest):
        thesplog('Handling %s actor exit request', 'recursive' if exitRequest.isRecursive() else 'non-recursive', level=logging.CRITICAL)
        children = list(self._addrManager.sendingAddresses())

        if exitRequest.isRecursive() and children:
            for each in children:
                self._send_intent(TransmitIntent(each, exitRequest,
                                                 onError=lambda m: self._childInaccessible(each, m)))

        else:
            thesplog('Handling %s actor exit request: scheduling child actor exited to parent', 'recursive' if exitRequest.isRecursive() else 'non-recursive', level=logging.CRITICAL)
            self._sayGoodbye()

        thesplog('Handling %s actor exit request: setting _exiting', 'recursive' if exitRequest.isRecursive() else 'non-recursive', level=logging.CRITICAL)
        self._exiting = True  # set exiting mode

#KWQ: move on to testDeadLettering
        if exitRequest.isRecursive() and children:
            # Should wait for children to exit before this Actor exits, but drop everything else
            thesplog('Handling %s actor exit request: wait for children', 'recursive' if exitRequest.isRecursive() else 'non-recursive', level=logging.CRITICAL)
            return True  # keep going
        # Don't need to wait for children, so exit as soon as transmit pipe drains.
        thesplog('Handling %s actor exit request: drain transmits', 'recursive' if exitRequest.isRecursive() else 'non-recursive', level=logging.CRITICAL)
        self.transport.abort_run(drain=True)
        return True


    def _sayGoodbye(self):
        self._send_intent(
            TransmitIntent(self._parentAddr, ChildActorExited(self.transport.myAddress)))


    def _childInaccessible(self, childAddress, exitRequestIntent):
        thesplog('_childInaccessible %s', childAddress)
        self._handleChildExited(childAddress)


    # ----------------------------------------------------------------------
    # Transmit management

    def actor_send(self, targetAddr, msg):
        self._send_intent(TransmitIntent(targetAddr, msg,
                                         onSuccess=self._checkNextTransmit,
                                         onError = self._checkNextTransmit))


    # ----------------------------------------------------------------------
    # Child Actor Management

    def createActor(self, newActorClass, targetActorRequirements, globalName,
                    sourceHash = None):
        naa = self._addrManager.createLocalAddress()
        if not globalName:
            try:
                self._startChildActor(naa, newActorClass,
                                      # KWQ: childRequirements?
                                      sourceHash = sourceHash or self._sourceHash)
                # transport will contrive to call _pendingActorReady when the
                # child is initialized and connected to this parent.
                return naa
            except NoCompatibleSystemForActor:
                pass

        self.transport.scheduleTransmit(
            TransmitIntent(self._adminAddr,
                           PendingActor(newActorClass,
                                        self.transport.myAddress, naa.addressDetails.localActorNum,
                                        targetActorRequirements,
                                        globalName=globalName,
                                        sourceHash=sourceHash)))
        return naa


    def _pendingActorReady(self, childInstance, actualAddress):
        thesplog('_pendingActorReady %s %s', childInstance, actualAddress)
        self._addrManager.registerUseableAddress(actualAddress)
        self._addrManager.associateUseableAddress(self.myAddress, childInstance, actualAddress)

        # Send any queued transmits for this child, and move the
        # finalTransmit marker to use the new address.

        lcladdr = self._addrManager.getLocalAddress(childInstance)

        if lcladdr in self._finalTransmitPending:
            self._finalTransmitPending[actualAddress] = self._finalTransmitPending[lcladdr]
            del self._finalTransmitPending[lcladdr]

        if lcladdr in self._awaitingAddressUpdate:
            pending = self._awaitingAddressUpdate[lcladdr]
            del self._awaitingAddressUpdate[lcladdr]
            thesplog('_pendingActorReady %s (#%s) %s has %s pending transmits', lcladdr, childInstance, actualAddress, len(pending))
            for each in pending:
                if each.targetAddr == lcladdr:
                    each.changeTargetAddr(actualAddress)
                thesplog('_retrying %s to %s', each.message, each.targetAddr)
                # The only way this intent would have been on an
                # _awaitingAddressUpdate list is if a previous attempt
                # to actually transmit it it had already called
                # _send_intent_to_transport; preservation of transmit
                # ordering dictates it's still the right intent to
                # perform:
                self._send_intent_to_transport(each)
        else: thesplog('_pendingActorReady %s %s no pending transmits for %s (%s): %s hashes %s', childInstance, actualAddress, lcladdr, hash(lcladdr), str(self._awaitingAddressUpdate), str([hash(k) for k in self._awaitingAddressUpdate]))

    def _pendingActorResponse(self, envelope):
        thesplog('_pendingActorResponse %s', envelope)
        if envelope.message.isSuccess:
            self._pendingActorReady(envelope.message.childInstance, envelope.message.actualAddress)
            return True
        # Pending Actor Creation failed, clean up all the stuff associated with the intended Actor
        lcladdr = self._addrManager.getLocalAddress(self.instanceNum)
        self._receiveQueue.append(ChildActorExited(lcladdr))
        if lcladdr in self._awaitingAddressUpdate:
            for each in self._awaitingAddressUpdate[lcladdr]:
                if not isinstance(each.message, TellMessage):
                    self._receiveQueue.append(PoisonMessage(each.message))
            del self._awaitingAddressUpdate[lcladdr]


    """
    def getActorStatus(self, statusReq, sender):
        resp = Thespian_ActorStatus(self.address,
                                    self._instance.__class__.__name__,
                                    self._adminAddress)
        for C in self._children:
            resp.addChild(self._children[C].childAddress,
                          -1 if self._children[C].childProcess is None else self._children[C].childProcess.pid)
        for F in self._msgsPending:
            resp.addPendingMessage(F[0], str(F[1]))
        self._sendOrQueue(sender, resp)
    """
