import types
from thespian.actors import ActorAddress


class ActorLocalAddress:
    """Used as the detail for an ActorAddress where the ActorAddress was
       generated locally by an Actor for a created child Actor in lieu
       of the actual final address of that child Actor.  This object
       defines the target ActorAddress in context of the Actor which
       generated that address, and is only valid for use by that
       generating Actor.  As such, it is not itself pickleable
       (although it may have a useable equivalent that can be
       substituted when pickling).

       The exception is that the address may be sent to the Admin for
       purposes of Creating or Forwarding to the actual destination
       actor.  For this purpose, the ActorLocalAddressExportedToAdmin
       object should be used, as generated from this object.

    """
    def __init__(self, generatingActorAddress, addressInstanceNum, addrManager):
        self.generatingActor    = generatingActorAddress
        self.addressInstanceNum = addressInstanceNum
        self.addressManager     = addrManager
    def __eq__(self, o):
        # n.b. compareAddressEq below expects this to throw an exception if o is not an ActorLocalAddress
        return self.generatingActor == o.generatingActor and self.addressInstanceNum == o.addressInstanceNum
    def __hash__(self):  return self.addressInstanceNum + hash(self.generatingActor)
    def __str__(self): return 'LocalAddr.%s'%self.addressInstanceNum


class CannotPickle(Exception):
    """This exception is thrown for objects that explicitly reject the
       effort to pickle them for sending them to a remote Actor."""
    pass


class CannotPickleAddress(Exception):
    """This exception is thrown for unpickleable ActorAddress objects.
       This is usually because it's a local address without, as yet, a
       useable translation.  This send might be retried later when the
       remote Actor Address is known.
    """
    def __init__(self, address, *args, **kw):
        super(CannotPickleAddress, self).__init__(*args, **kw)
        self.address = address


class ActorLocalAddressExportedToAdmin:
    """This represents an ActorLocalAddress in a form that should be
       exported ONLY to the ActorSystem administration."""
    def __init__(self, actorLocalAddress):
        self.creatingActor   = actorLocalAddress.generatingActor
        self.creationInstNum = actorLocalAddress.addressInstanceNum
    def __str__(self): return 'AdminExportedAddress(src=%s, #%s)'%(str(self.creatingActor), self.creationInstNum)


def _pickle_if_translation(lclAddr):
    """Assigned to ActorAddresses of ActorLocalAddress details.  Prevents
       pickling unless there is a useable translation, in which case
       that translation is the pickled version.
    """
    useableAddr = lclAddr.addressDetails.addressManager.sendingAddress(lclAddr)
    if useableAddr:
        return useableAddr.__getstate__() \
            if hasattr(useableAddr, '__getstate__') else useableAddr.__dict__
    raise CannotPickleAddress(lclAddr, 'ActorLocalAddress %s cannot be Pickled.', str(lclAddr))


def _pickle_clean(dirtyAddr):
    "Cleans up an ActorAddress of any special, unpickleable elements so it can be pickled."
    cd = dirtyAddr.__dict__.copy()
    for rmv in [ 'eqOverride', '__getstate__', '_pickle_clean']:
        if rmv in cd:
            del cd[rmv]
    return cd


class ActorAddressManager:
    """Class used to manage ActorLocalAddress addresses and their
       translation to a directly-useable ActorAddress."""
    def __init__(self, thisActorAddress):
        self._myActor = thisActorAddress  # managing address on behalf of this Actor
        self._managed = []  # value = None or directly-useable Address
        self._deadAddrs = [] # value = dead ActorAddress

    def createLocalAddress(self):
        """Creates a new ActorLocalAddress-based ActorAddress.  A useable
           Address will be associated with this address in the future."""
        self._managed.append(None)
        return self.getLocalAddress(len(self._managed) - 1)

    def getLocalAddress(self, instanceNum):
        'Returns ActorAddress of local address instance; throws exception on invalid instanceNum'
        ra =  ActorAddress(ActorLocalAddress(self._myActor, instanceNum, self))
        ra.eqOverride = types.MethodType(self.compareAddressEq, ra)
        ra.__getstate__ = types.MethodType(_pickle_if_translation, ra)
        return ra

    def registerUseableAddress(self, anAddr):
        """Called for any useable Address being passed to the Actor; ensures
           that any comparison between this useable Address and a
           local Address created originally will be properly
           performed.
        """
        anAddr.eqOverride = types.MethodType(self.compareAddressEq, anAddr)
        anAddr.__getstate__ = types.MethodType(_pickle_clean, anAddr)

    def reserveLocalAddress(self, instanceNum):
        while len(self._managed) < instanceNum + 1:
            self._managed.append(None)
        # if instanceNum != len(self._managed):
        #     raise ValueError('cannot reserve arbitrary instances (%d != %d)'%(instanceNum,
        #                                                                       len(self._managed)))
        # self._managed.append(None)

    def deadAddress(self, address):
        self._deadAddrs.append(address)
        self._managed = filter(lambda a: a != address, self._managed)

    def isDeadAddress(self, address):
        return address in self._deadAddrs

    def associateUseableAddress(self, ownerAddress, ownerInstance, useableAddress):
        """Called when the actual Actor Address becomes known for a
           locally-generated Address (e.g. when the remote Actor has
           been created and addressed).  Associates the local address
           with the remote address for future use.
        """
        if ownerAddress == self._myActor:
            self._managed[ownerInstance] = useableAddress

    def compareAddressEq(self, addr1, addr2):
        "Checks if two addresses are equal, considering local/useable associations"
        print('compareAddressEq %s %s'%(addr1, addr2))
        if addr1 is None: return addr2 is None
        if addr2 is None: return False
        try:
            return addr1.addressDetails == addr2.addressDetails
        except AttributeError:
            # Cannot directly compare address details elements; try more sophisticated checks
            pass
        if isinstance(addr1.addressDetails, ActorLocalAddress):
            try:
                return (addr1.addressDetails.generatingActor == self._myActor and
                        self._managed[addr1.addressDetails.addressInstanceNum] and
                        self._managed[addr1.addressDetails.addressInstanceNum].addressDetails ==
                        addr2.addressDetails)
            except AttributeError:
                # Cannot compare these details... must not be equivalent addresses
                return False
        if isinstance(addr2.addressDetails, ActorLocalAddress):
            try:
                return (addr2.addressDetails.generatingActor == self._myActor and
                        self._managed[addr2.addressDetails.addressInstanceNum] and
                        self._managed[addr2.addressDetails.addressInstanceNum].addressDetails ==
                        addr1.addressDetails)
            except AttributeError:
                # Cannot compare these details... must not be equivalent addresses
                return False
        # IncomPARable!!

    def sendingAddress(self, anAddress):
        "Returns an ActorAddress useable for sending messages, or None if not available yet."
        if isinstance(anAddress.addressDetails, ActorLocalAddress):
            if anAddress.addressDetails.generatingActor == self._myActor:
                return self._managed[anAddress.addressDetails.addressInstanceNum]
            return None
        elif isinstance(anAddress.addressDetails, ActorLocalAddressExportedToAdmin):
            if anAddress.addressDetails.creatingActor == self._myActor:
                try:
                    return self._managed[anAddress.addressDetails.creationInstNum]
                except IndexError:
                    return None  # Not yet setup
            return None
        else:
            # Assumed to be directly useable
            return anAddress

    def sendingAddresses(self):
        """Iterates through all established useable sending addresses as a
           generator.  Ignores still unresolved sending addresses."""
        for addr in filter(None, self._managed): yield addr

    @staticmethod
    def exportAddress(actorAddr):
        if isinstance(actorAddr.addressDetails, ActorLocalAddress):
            return ActorAddress(ActorLocalAddressExportedToAdmin(actorAddr.addressDetails))
        return actorAddr
