"""Common functionality for multiprocess system basees built with the
python 'multiprocess' module.  Intended as a base class, not for
direct usage."""


import logging
from thespian.actors import *
from thespian.system.systemBase import systemBase, thesplog
from thespian.system.transport import *
import multiprocessing
from datetime import timedelta


MAX_ADMIN_STARTUP_DELAY = timedelta(seconds=5)


def setProcName(name, actorAddr):
    try: from setproctitle import setproctitle
    #This library not required, but its presence will make
    #actor names and addresses available in the process list.
    except: pass
    else: setproctitle('%s %s'%(name, str(actorAddr)))


#KWQ: rename systemBase to systemInterface, because it only exists for external process talking into and ActorSystem (that might exist independently)

class multiprocessCommon(systemBase):

    def __init__(self, system, logDefs = None):
        self.transport = self.transportType(ExternalInterfaceTransportInit(),
                                            system.capabilities, logDefs)
        super(multiprocessCommon, self).__init__(system, logDefs)


    def _startAdmin(self, adminAddr, addrOfStarter, capabilities, logDefs):
        endpointPrep = self.transport.prepEndpoint(adminAddr)

        multiprocessing.process._current_process._daemonic = False
        admin = multiprocessing.Process(target=startAdmin,
                                        args=(MultiProcAdmin,
                                              addrOfStarter,
                                              endpointPrep,
                                              self.transport.__class__,
                                              adminAddr,
                                              capabilities,
                                              logDefs),
                                        name='ThespianAdmin')
        admin.daemon = True
        #multiprocessing.process._current_process._daemonic = True
        admin.start()
        # admin must be explicity shutdown and is not automatically
        # stopped when this current process exits.
        multiprocessing.process._current_process._children.remove(admin)

        self.transport.connectEndpoint(endpointPrep)

        response = self.transport.run(None, MAX_ADMIN_STARTUP_DELAY)
        if not response or not isinstance(response.message, EndpointConnected):
            raise InvalidActorAddress(adminAddr, 'ActorSystem admin did not start')



from thespian.system.systemAdmin import AdminCore

def startAdmin(adminClass, addrOfStarter, *args):
    # Slight trickiness here.  There may *already* be an admin bound
    # to this start address.  However, the external process attempting
    # to start is going to wait for the EndpointConnected message
    # before continuing, so ensure that message is *not* sent until
    # the local admin Transport has had time to bind and listen to the
    # local address, but also ensure that the message is *always* sent
    # even if the local admin could not start (caller will use
    # _verifyAdminRunning to ensure things are OK.
    try:
        admin = adminClass(*args)
    except Exception:
        #admin.transport.scheduleTransmit(TransmitIntent(addrOfStarter, EndpointConnected(0)))
        thesplog('KWQ Need to have explicit establishEndpoint in child instead of binding into class startup so that the endpoint can be checked even on failure (as above)... or ekse move the endpoint connected acknowledgement into the transport?', level=logging.DEBUG)
        raise
    admin.transport.scheduleTransmit(TransmitIntent(addrOfStarter, EndpointConnected(0)))
    admin.run()


class MultiProcReplicator(object):

    def _startChildActor(self, childAddr, childClass,
                         childRequirements=None,
                         sourceHash=None):
        # KWQ: when child starts it will have this parent address and it will initialize its transport and notify the parent, whereupon the parent will see the incoming message from the child with the id# indicated in the addressmanager localaddress and update the localaddress.  All this should happen in the transport though, not here.
        endpointPrep = self.transport.prepEndpoint(childAddr)

        multiprocessing.process._current_process._daemonic = False
        child = multiprocessing.Process(target=startChild,  #KWQ: instantiates module specified by sourceHash to create actor
                                        args=(childClass,
                                              endpointPrep,
                                              self.transport.__class__,
                                              sourceHash or self._sourceHash,
                                              self.myAddress,
                                              self._adminAddr,
                                              childRequirements),
                                        name='Actor_%s__%s'%(childClass, str(childAddr)))
        child.daemon = True
        #multiprocessing.process._current_process._daemonic = True
        child.start()
        # Also note that while non-daemonic children cause the current
        # process to automatically join() those children on exit,
        # daemonic children are sent a terminate() operation (usually
        # indicated by a SIGTERM under unix or TERMINATE indicator
        # under windows.  To avoid this, use another dirty trick and
        # remove all children from the _current_process._children list
        # so that they are not automatically stopped when this process
        # stops.
        multiprocessing.process._current_process._children.remove(child)

        if not hasattr(self, '_child_procs'): self._child_procs = []
        self._child_procs.append(child)
        thesplog('adding new child %s (%s) @ %s to childprocs', child, childClass, childAddr, level=logging.CRITICAL)

        self.transport.connectEndpoint(endpointPrep)


    @staticmethod
    def _checkChildLiveness(child):
        if not child.is_alive():
            # Don't join forever; that might hang and it's ok to leave
            # zombies as long as we continue to make progress.
            child.join(0.5)
            return False
        return True

    def _childExited(self, childAddr):
        thesplog('child @ %s exited, checking join', childAddr, level=logging.CRITICAL)
        # KWQ: this finds children that may not have sent the ChildActorExited yet.... should probably try to correlate.  Could also use this as a way to detect failed actors, but that should give them some time to send their ChildActorExited and then synthesize one if it's been too long.
        self._child_procs = [C for C in self._child_procs if self._checkChildLiveness(C)]
        thesplog('%s child procs left: %s', len(self._child_procs), map(lambda c: '#%d:%s'%(c.pid, str(c)), self._child_procs), level=logging.CRITICAL)


    def h_EndpointConnected(self, envelope):
        self._pendingActorReady(envelope.message.childInstance, envelope.sender)
        return True

    def _handleReplicatorMessages(self, envelope):
        if isinstance(envelope.message, EndpointConnected):
            return True, self.h_EndpointConnected(envelope)
        return False, True



class EndpointConnected(object): # internal message sent from child to parent on startup connection
    def __init__(self, childInstance):
        self.childInstance = childInstance


from thespian.system.actorManager import ActorManager

def startChild(childClass, endpoint, transportClass, sourceHash, parentAddr, adminAddr,
               childRequirements):
    # Dirty trick here to workaround multiprocessing trying to impose
    # an unnecessary restriction.  A process should be set daemonic
    # before start() if the parent shouldn't track it (an specifically
    # automatically join() the subprocess on exit).  For Actors, the
    # parent exists independently of the child and the ActorSystem
    # manages them, so daemonic processes are desired.  However,
    # multiprocessing imposes a restriction that daemonic processes
    # cannot create more processes.  The following reaches deep into
    # the implementation of the multiprocessing module to override
    # that restriction.  This process was already started as daemonic,
    # and it's detached from its parent.  The following simply clears
    # that flag locally so that other processes can be created from
    # this one.
    multiprocessing.process._current_process._daemonic = False

    am = MultiProcManager(childClass, endpoint, transportClass, sourceHash, parentAddr, adminAddr,
                          childRequirements)
    am.transport.scheduleTransmit(TransmitIntent(parentAddr, EndpointConnected(endpoint.addrInst)))
    setProcName(childClass.__name__, am.transport.myAddress)
    am.run()



class MultiProcAdmin(MultiProcReplicator, AdminCore): pass
class MultiProcManager(MultiProcReplicator, ActorManager): pass

