'''The multiprocessSystemBase runs each Actor in its own separate
process for a truly asynchronous system.  The processes can
communicate using network channels.
'''

import logging, logging.handlers
try:
    from logging.config import dictConfig
except ImportError:
    # Old python that doesn't contain this...
    from thespian.system.dictconfig import dictConfig

import string, types, socket, pickle, errno
import traceback
import datetime, time
import multiprocessing
from thespian.actors import *
from thespian.system import *
from thespian.system.transport.socketAddressMgmt import *
from thespian.system.messages.admin import *


import datetime
def mlog(msg):
    with open('/tmp/mproc.log','a') as lf: lf.write('%s %s\n'%(str(datetime.datetime.now()), str(msg)))


PENDING_SHUTDOWN_RETRIES          = 50
PENDING_SHUTDOWN_DELAY            = 0.25 # seconds
MAX_ACTOR_STARTUP_RETRIES         = 60
CONVENTION_REREGISTRATION_PERIOD  = datetime.timedelta(minutes=4, seconds=22)
CONVENTION_RESTART_PERIOD         = datetime.timedelta(seconds=22)
CONVENTION_REGISTRATION_TIMEOUT   = 900  # 15 minutes
CONVENTION_REGISTRATION_MISS_MAX  = 3  # # of missing convention registrations before death declared
CREATE_ACTOR_THROTTLE_DELAY       = 0.01  # seconds; allows new actor to get started
ADMIN_SEND_TIMEOUT                = 0.1
PENDING_LOG_SEND_PERIOD           = 0.025 # seconds to wait between attempting to send to logger when blocked
LOG_SEND_TIMEOUT                  = 0.25 # seconds to try (or between retries) forwarding logs to convention leader
SEND_QUEUE_RETRY_PERIOD           = 0.023 # seconds between queued send retries
RETRY_DELAY_INCREASE              = 1.33
MAX_RETRY_DELAY                   = 1.25

def timePeriodSeconds(basis, other=None):
    if isinstance(basis, datetime.datetime):
        if isinstance(other, datetime.datetime):
            return timePeriodSeconds(other - basis)
    if isinstance(basis, datetime.timedelta):
        try:
            return basis.total_seconds()
        except AttributeError:
            # Must be Python 2.6... which doesn't have total_seconds yet
            return (basis.days * 24.0 * 60 * 60) + basis.seconds + (basis.microseconds / 1000.0 / 1000)
    raise TypeError('Cannot determine time from a %s argument'%str(type(basis)))


def setProcName(name, actorAddr):
    try: from setproctitle import setproctitle
    #This library not required, but its presence will make
    #actor names and addresses available in the process list.
    except: pass
    else: setproctitle('%s %s'%(name, str(actorAddr)))

# ----------------------------------------------------------------------
# ActorSystemAdmin


class DeadLetterAddress(object):
    """Message sent to the Admin when it is to assume DeadLetterBox
    responsibility for the specified address.
    """
    def __init__(self, deadAddr, why='unknown'):
        self.deadAddress = deadAddr
        self.why         = why


class HandleDeadLetters(object):
    """Message sent to the Admin to register or de-register the specified
       address for DeadLetterBox handling."""
    def __init__(self, handlerAddr, enableHandler):
        self.handlerAddr   = handlerAddr
        self.enableHandler = enableHandler


class NotifyOnSystemRegistration(object):
    """Message sent to the Admin to register or de-register the specified
       address for ActorSystem Convention Registration handling."""
    def __init__(self, handlerAddr, enableNotification):
        self.handlerAddr        = handlerAddr
        self.enableNotification = enableNotification


class ForwardMessage(object):
    """Message sent from the external ActorSystem to the Admin wrapping a
       message that should be sent to a (possibly remote) Actor once
       that Actor is available.
    """
    def __init__(self, toActorAddress, msg):
        self.toActorAddress = toActorAddress  # details probably ActorLocalAddressExportedToAdmin
        self.msg     = msg


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# Special Logging pseudo-Actor
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------

# This is the process that handles logging output for all the Actors
# in this ActorSystem (and possibly forwarded logs from other
# ActorSystems).  This is similar to an Actor, and responds to
# messages like an Actor, but the startup of this is slightly
# different in that it starts without knowledge of the Admin and by
# special initialization.  It will process messages like other Actors,
# but will never talk back to the Admin (which it doesn't know about).
# It will ignore normal ActorExitRequest messages as well (in case it
# gets a spurious one) and only exit on receiving a LoggerExitRequest,
# which should only be generated by the local Admin.

class LoggerExitRequest(object): pass
class LoggerFileDup(object): pass

def startupASLogger(actorAddress, logSocket, logDefs, conventionAddress, closeList):
    # Dirty trick here to completely re-initialize logging in this
    # process... something the standard Python logging interface does
    # not allow via the API.  We also do not want to run
    # logging.shutdown() because (a) that does not do enough to reset,
    # and (b) it shuts down handlers, but we want to leave the parent's
    # handlers alone.
    # Dirty trick here to completely re-initialize logging in this
    # process... something the standard Python logging interface does
    # not allow via the API.
    logging.root = logging.RootLogger(logging.WARNING)
    logging.Logger.root = logging.root
    logging.Logger.manager = logging.Manager(logging.Logger.root)
    if logDefs:
        dictConfig(logDefs)
    else:
        logging.basicConfig()
    logging.info('ActorSystem Logging Initialized')
    mlog('ActorSystem Logging Initialized')
    setProcName('logger', actorAddress)
    import select
    queued = []
    fdup = None
    try:
        while True:
            (rdy, ig1, ig2) = select.select([logSocket], [], [],
                                            PENDING_LOG_SEND_PERIOD if queued else None)
            for rs in rdy:
                try:
                    msg = acceptAndReceive(rs, None)
                    if msg:
                        logsender, logrecord = msg
                        if isinstance(logrecord, LoggerExitRequest):
                            raise FloatingPointError('special exit signal')
                        elif isinstance(logrecord, LoggerFileDup):
                            fdup = getattr(logrecord, 'fname', None)
                        elif isinstance(logrecord, logging.LogRecord):
                            logging.getLogger(logrecord.name).handle(logrecord)
                            if fdup:
                                with open(fdup, 'a') as ldf: ldf.write('%s\n'%str(logrecord))
                            mlog(str(logrecord))
                            if conventionAddress and \
                               logrecord.levelno >= logging.WARNING:
                                if sendToAddress(conventionAddress,
                                                 logrecord,
                                                 actorAddress,
                                                 LOG_SEND_TIMEOUT) == SendStatus.NotSent:
                                    queued.append(logrecord)
                        else:
                            mlog('Unknown message rcvd by logger: %s'%str(msg))
                            logging.warn('Unknown message rcvd by logger: %s'%str(msg))
                except Exception as ex:
                    mlog('Exception receiving log message: %s'%str(ex))
                    logging.warn('Exception receiving log message', exc_info=True)
            while queued:
                qued = queued.pop(0)
                if sendToAddress(conventionAddress,
                                 qued,
                                 actorAddress,
                                 LOG_SEND_TIMEOUT) == SendStatus.NotSent:
                    queued.insert(0, qued)
                    break
    except FloatingPointError:
        logging.info('ActorSystem Logging Shutdown')
    except:
        logging.error('Error selecting for logging', exc_info=True)
    finally:
        logSocket.close()


def startASLogger(logDefs, conventionAddress=None, closeList=[]):
    for H in closeList:
        if H:
            try:
                H.close()
            except:
                pass

    logAddr, logSocket = newActorAddress(name='ActorLogger')
    multiprocessing.process._current_process._daemonic = False
    logProc = multiprocessing.Process(target=startupASLogger,
                                      args = (logAddr, logSocket, logDefs,
                                              conventionAddress, closeList))
    logProc.daemon = True
    logProc.start()
    time.sleep(0.01)  # pause for logger to start
    logSocket.close() # still open in logging process
    return logAddr


class ThespianLogForwardHandler(logging.Handler):
    def __init__(self, toAddr, fromAddr):
        logging.Handler.__init__(self, 1)  # forward EVERYTHING
        self._name = 'ThespianLogForwardHandler'
        self._fwdAddr = toAddr
        self._fromAddr = fromAddr
    def handle(self, record):
        # Can't pickle traceback objects, so pre-format them.  Sorry, more logging internals here.
        if record.exc_info:
            if not record.exc_text:
                excinfo = traceback.format_exception(record.exc_info[0], record.exc_info[1],
                                                     record.exc_info[2])
                record.exc_text = '\n'.join(excinfo)
            record.exc_info = None
        record.__dict__['actorAddress'] = str(self._fromAddr)
        msg = record.getMessage()
        record.msg = msg
        record.args = None
        for retries in range(4):
            if sendToAddress(self._fwdAddr, record,
                             self._fromAddr,
                             LOG_SEND_TIMEOUT) != SendStatus.NotSent:
                return
        # four failures... drop it.


class ThespianLogForwarder(logging.Logger):
    def __init__(self, toAddr, fromAddr):
        # N.B.  In Python2.6, Logger and Handler don't derive from
        # object, so super() cannot be used here.
        logging.Logger.__init__(self, 'root', 1)  # forward EVERYTHING
        logging.Logger.addHandler(self, ThespianLogForwardHandler(toAddr, fromAddr))
    def addHandler(self, hdlr):
        raise NotImplementedError('Cannot add logging handlers for Thespian Actors')
    def removeHandler(self, hdlr):
        raise NotImplementedError('Cannot add logging handlers for Thespian Actors')



# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# Special ActorSystem shutdown pending process and message
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------


class PendingShutdown(object):
    def __init__(self, cnt, requester):
        self.cnt = cnt
        self.requester = requester


# n.b. Not an actual Actor because this is used during shutdown time
# when the ActorSystem is providing normal Actor management
# (e.g. ChildActorExited and DeadLetterAddress handling).

def notifyAfterTimeout(timeout_period, adminAddr, timeoutMsg):
    tproc = multiprocessing.Process(target=start_TProc,
                                    args = (timeout_period,
                                            adminAddr,
                                            timeoutMsg))
    tproc.start()

def start_TProc(timeout_period, adminAddr, timeoutMsg):
    import time
    time.sleep(timeout_period)
    attempts = 10
    while attempts and sendToAddress(adminAddr, timeoutMsg, adminAddr) == SendStatus.NotSent:
        attempts = attempts - 1
        time.sleep(0.1)
    # exits


# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ActorSystem Admin
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------


class ConventionMember(object):
    def __init__(self, address, capabilities):
        self.remoteAddress      = address
        self.remoteCapabilities = capabilities
        self.lastVerifiedAt     = datetime.datetime.now()
        self.hasRemoteActors    = []  # (localParent, remoteActor) addresses created remotely
    def createdActor(self, localParentAddress, newActorAddress):
        entry = localParentAddress, newActorAddress
        if entry not in self.hasRemoteActors:
            self.hasRemoteActors.append(entry)
    def __str__(self):
        return 'ActorSystem @ %s on %s'%(str(self.remoteAddress), str(self.lastVerifiedAt))




class ActorSystemAdmin(object):

    '''This is the Administrator, which is the central coordinating
    process for this ActorSystem.  One of the primary purposes of the
    ActorSystemAdmin is providing the DeadLetterBox for
    terminated Actors.
    '''

    def __init__(self, adminAddress, adminSock, system, logger):
        self.address   = adminAddress
        self.adminSock = adminSock
        self.system    = system
        self.logger    = logging.getLogger('Thespian.Admin')
        self.loggingActor = logger
        self.deadZone  = []  # array of (addr,sock)
        self.deadLetterHandler = None
        self.conventionNotificationHandler = None
        self.pendingMessages = []  # (fromActor, ForwardMessage)
        self.conventionMembers = {} # key=ActorSystem ActorAddress string, value=ConventionMember
        self.addressManagers = {} # key=ActorAddressString of actor originally creating address
                                  # value=ActorAddressManager(self._address)
        self.systemRmtActorSender = {} # key=Local Index, value=Sender (replyTo)
        self._globalNames = {} # key=globalName, value=ActorAddress (never None)
        self._queuedsends = []   # (rmtaddress, msg, fromaddr, timeseq, last_attempt_time)
        self._conventionLeaderIsGone = True
        self._conventionFirstTime = True
        self._conventionLastRegisteredAt = datetime.datetime(1970,1,1)
        self.setupConvention()


    def _sendOne(self, toAddr, msg, fromAddr):
        if toAddr in [dead[0] for dead in self.deadZone]:
            if self.deadLetterHandler:
                sts = sendToAddress(self.deadLetterHandler, msg, fromAddr, ADMIN_SEND_TIMEOUT)
            else:
                return SendStatus.Sent # drop the message
        else:
            sts = sendToAddress(toAddr, msg, fromAddr, ADMIN_SEND_TIMEOUT)
        if sts == SendStatus.BadPacket:
            # KWQ: log this
            sts = sendToAddress(fromAddr, PoisonMessage(msg), toAddr, ADMIN_SEND_TIMEOUT)
            if sts == SendStatus.BadPacket:
                # show never happen
                # KWQ: log this
                return SendStatus.Sent
        return sts


    def _sendOrQueue(self, rmtAddress, msg, fromActor=None):
        if fromActor is None:
            fromActor = self.address
        if not self._queuedsends:
            if self._sendOne(rmtAddress, msg, fromActor) == SendStatus.Sent:
                return
        self._queuedsends.append( (rmtAddress, msg, fromActor, list(timeseq(300, origDelay=0.37)), datetime.datetime.now() ) )
        # self._sendQueued()


    def _sendQueued(self):
        failedTo = []
        queued = self._queuedsends
        self._queuedsends = []
        now = datetime.datetime.now()
        for tgt,mesg,frm,delays,lastTried in queued:
            if tgt not in failedTo:
                elapsed = (((now - lastTried) * 100).seconds * 1.0) / 100.0
                if elapsed >= delays[0]:
                    if self._sendOne(tgt, mesg, frm) == SendStatus.Sent:
                        continue
                    delays = delays[1:]
                    lastTried = now
                failedTo.append(tgt)
            if delays:
                self._queuedsends.append( (tgt,mesg,frm,delays,lastTried) )
            else:
                self.logger.warning('Dropped send of %s from %s to %s; could not send in a reasonable time period.',
                                    str(mesg), str(frm), str(tgt))


    def setupConvention(self):
        actorConvention = self.system.capabilities.get('Convention Address.IPv4', None)
        if actorConvention:
            self.conventionAddress = ActorAddress(SocketActorAddress(socket.AF_INET,
                                                                     socket.SOCK_STREAM,
                                                                     socket.IPPROTO_TCP,
                                                                     actorConvention))
            self.logger.info('Admin registering with Convention @ %s'%str(self.conventionAddress))
            try:
                sendOrFail(self.conventionAddress,
                           ConventionRegister(self.address, self.system.capabilities,
                                              self._conventionFirstTime),
                           self.address, CONVENTION_REGISTRATION_TIMEOUT)
                self._conventionLastRegisteredAt = datetime.datetime.now()
                self._conventionLeaderIsGone     = False
                self._conventionFirstTime        = False
            except SendTimeoutError:
                self.logger.warning('Admin cannot register with Convention @ %s',
                                    str(self.conventionAddress))
                self._conventionLeaderIsGone = True
            except SendStatus.BadPacketError:
                self.logger.warning('Convention @ %s incompatible with Admin',
                                    str(self.conventionAddress))
                self._conventionLeaderIsGone = True
        else:
            self.conventionAddress = None  # still might be a Convention Leader


    def isConventionLeader(self): return self.conventionAddress is None


    def handleConventionRegistration(self, msg, sender):
        # Registrant may re-register if changing capabilities
        registrant = msg.adminAddress
        self.logger.info('Got Convention registration from %s', str(registrant))
        #erase our knowledge of actors associated with potential former instance of this system
        if msg.firstTime: self._remoteSystemCleanup(registrant)
        newReg = registrant.actorAddressString not in self.conventionMembers
        self.conventionMembers[registrant.actorAddressString] = ConventionMember(registrant,
                                                                                 msg.capabilities)
        if newReg and self.conventionNotificationHandler:
            self._sendOrQueue(self.conventionNotificationHandler,
                              ActorSystemConventionUpdate(registrant, msg.capabilities, True))
        # If we are the Convention Leader, inform all other registrants of the new registrant
        if self.isConventionLeader():
            # Inform new registrant of our capabilities
            self._sendOrQueue(sender,
                              ConventionRegister(self.address, self.system.capabilities))
            # # Inform new registrant of other registrants
            # if len(self.conventionMembers) > 1:
            #     self.logger.debug('Notifying other Convention members of registration'
            #                       'from %s'%str(registrant))
            # for member in self.conventionMembers:
            #     if self.conventionMembers[member].remoteAddress != sender:
            #         self._sendOrQueue(self.conventionMembers[member].remoteAddress, msg)


    def handleConventionDeRegistration(self, msg, sender):
        self._remoteSystemCleanup(msg.adminAddress)

    def _remoteSystemCleanup(self, registrant):
        """Called when a RemoteActorSystem has exited and all associated
           Actors should be marked as exited and the ActorSystem
           removed from Convention membership.
        """
        if registrant.actorAddressString in self.conventionMembers:
            cmr = self.conventionMembers[registrant.actorAddressString]

            # Send exited notification to conventionNotificationHandler (if any)
            if self.conventionNotificationHandler:
                self._sendOrQueue(self.conventionNotificationHandler,
                                  ActorSystemConventionUpdate(cmr.remoteAddress,
                                                              cmr.remoteCapabilities, False))

            # If the remote ActorSystem shutdown gracefully (i.e. sent
            # a Convention Deregistration) then it should not be
            # necessary to shutdown remote Actors (or notify of their
            # shutdown) because the remote ActorSystem should already
            # have caused this to occur.  However, it won't hurt, and
            # it's necessary if the remote ActorSystem did not exit
            # gracefully.

            for lpa, raa in cmr.hasRemoteActors:
                self._sendOrQueue(lpa, ChildActorExited(raa))
                # n.b. at present, this means that the parent might
                # get duplicate notifications of ChildActorExited; it
                # is expected that Actors can handle this.

            # Remove remote system from conventionMembers
            del self.conventionMembers[registrant.actorAddressString]

        if registrant == self.conventionAddress:
            # Convention Leader has exited.  Do NOT set
            # conventionAddress to None.  It might speed up shutdown
            # of this ActorSystem because it won't try to de-register
            # from the convention leader, but if the convention leader
            # reappears there will be nothing to get this ActorSystem
            # re-registered with the convention.
            self._conventionLeaderIsGone = True


    def _checkConvention(self):
        if self.isConventionLeader():
            now = datetime.datetime.now()
            missingThreshold = CONVENTION_REREGISTRATION_PERIOD * CONVENTION_REGISTRATION_MISS_MAX
            missing = []
            for each in self.conventionMembers:
                if now - self.conventionMembers[each].lastVerifiedAt > missingThreshold:
                    missing.append(each)
            for each in missing:
                self.logger.warning('%s missed %d checkins (%s); assuming it has died',
                                    str(self.conventionMembers[each]),
                                    CONVENTION_REGISTRATION_MISS_MAX,
                                    str(missingThreshold))
                self._remoteSystemCleanup(self.conventionMembers[each].remoteAddress)
        else:
            # Re-register with the Convention if it's time
            if self.conventionAddress and \
               ((datetime.datetime.now() - self._conventionLastRegisteredAt) > self.checkin_period()):
                self.setupConvention()


    def handlePendingActor(self, pendingActorMsg, sender):
        """Internally called to startup an Actor that couldn't be directly
           started: generally because the capabilities of the local
           system didn't match or because it needs to register (or
           refer to) a globalName.  Examine known globalName Actors
           and then Convention members to see if there is a fit.
        """
        actorClass       = pendingActorMsg.actorClass
        requesterAddrStr = pendingActorMsg.forActor.actorAddressString
        globalName       = pendingActorMsg.globalName
        if requesterAddrStr not in self.addressManagers:
            self.addressManagers[requesterAddrStr] = ActorAddressManager(pendingActorMsg.forActor)
        self.addressManagers[requesterAddrStr].reserveLocalAddress(pendingActorMsg.instanceNum)

        if pendingActorMsg.forActor == self.address:
            self.systemRmtActorSender[pendingActorMsg.instanceNum] = sender


        cafr = CreateActorForRemote(actorClass,
                                    pendingActorMsg.forActor,
                                    pendingActorMsg.instanceNum,
                                    pendingActorMsg.targetActorReq,
                                    globalName, self.address)

        # If this is a known global name, satisfy caller's request
        # with the existing globalName-referenced Actor.

        if globalName:

            try:
                if globalName not in self._globalNames:
                    # If a globalName was specified, the requester deferred this
                    # to the Admin to see if there was an existing Actor.  At this
                    # point, there wasn't so try creating a local Actor to satisfy
                    # the request.

                    naa,nas = newActorAddress(name='%s Actor (%s)'%(
                        actorClass.__name__,
                        globalName or 'remotereq'))
                    try:
                        na = _createActor(actorClass, pendingActorMsg.targetActorReq,
                                          naa, nas,
                                          pendingActorMsg.forActor,
                                          self.address, self.loggingActor,
                                          systemCapabilities = self.system.capabilities)
                        if globalName:
                            self._globalNames[globalName] = naa
                    except NoCompatibleSystemForActor:
                        nas.close()
                        raise

                self.addressManagers[requesterAddrStr].associateUseableAddress(
                    pendingActorMsg.forActor,
                    pendingActorMsg.instanceNum,
                    self._globalNames[globalName])
                # n.b. forActor might be admin, in which case sender
                # is probably external askTo receiver; send response
                # back to sender, not forActor.
                # KWQ: use PendingActorResponse
                self._sendOrQueue(sender,
                                  RemoteActorCreated(cafr, self._globalNames[globalName]))
                return
            except NoCompatibleSystemForActor:
                # Couldn't create locally, so check other Convention members.
                pass

        for actorsys in self.conventionMembers:
            cm = self.conventionMembers[actorsys]
            if actorClass.actorSystemCapabilityCheck(cm.remoteCapabilities,
                                                     pendingActorMsg.targetActorReq):
                self._sendOrQueue(cm.remoteAddress, cafr)
                return

        if not self.isConventionLeader():
            self._sendOrQueue(self.conventionAddress, cafr)
            return

        # KWQ: add something to self.addressManagers[requesterAddrStr] to prevent future forwarding attempts?
        logging.warning('No matching system for creating actor %s'%str(actorClass))
        self._sendOrQueue(pendingActorMsg.forActor,
                          RemoteActorCreateFailed(cafr,
                                                  "No matching System for creating actor %s"%str(actorClass)))


    def createActorForRemote(self, createActorForRemoteMsg, sender):
        """Internally called when a remote ActorSystem in our Convention has
        requested an Actor to be created that it believes this
        ActorSystem can create.
        """
        actorClass = createActorForRemoteMsg.actorClass
        globalName = createActorForRemoteMsg.globalName

        if globalName and globalName in self._globalNames:
            self._sendOrQueue(createActorForRemoteMsg.adminAddr,
                              RemoteActorCreated(createActorForRemoteMsg,
                                                 self._globalNames[globalName]),
                              self.address)
            return

        naa,nas = newActorAddress(name='Remote Parented %s Actor'%(actorClass.__name__))
        try:
            na = _createActor(actorClass,
                              createActorForRemoteMsg.targetActorRequirements,
                              naa, nas,
                              createActorForRemoteMsg.forActor,
                              self.address, self.loggingActor,
                              closeList = [self.adminSock] + [dead[1] for dead in self.deadZone],
                              systemCapabilities = self.system.capabilities)
            # n.b. this actor is NOT managed by the local ActorSystem
            # Admin (i.e. it is not in the self._primaryActors
            # storage)...it is (already) managed by the remote Actor
            # that originated the creation of this Actor.  However,
            # any Dead-Lettering or subActor creation needed for this
            # new Actor will be handled by the local Admin.
            if globalName:
                self._globalNames[globalName] = naa
            self._sendOrQueue(createActorForRemoteMsg.adminAddr,
                              RemoteActorCreated(createActorForRemoteMsg, naa))  #KWQ: convert to external/remote version?
        except NoCompatibleSystemForActor:
            if self.isConventionLeader():
                for actorsys in self.conventionMembers:
                    cm = self.conventionMembers[actorsys]
                    if cm.remoteAddress != sender:
                        if actorClass.actorSystemCapabilityCheck(cm.remoteCapabilities,
                                                                 createActorForRemoteMsg.targetActorRequirements):
                            self._sendOrQueue(cm.remoteAddress, createActorForRemoteMsg)
                            return
            self._sendOrQueue(createActorForRemoteMsg.adminAddr,
                              #createActorForRemoteMsg.forActor,
                              #sender,
                              RemoteActorCreateFailed(createActorForRemoteMsg,
                                                      'Capabilities do not match'))
        except Exception as ex:
            mlog('Error creating actor %s for remote: %s', str(actorClass), str(ex))
            self.logger.warning('Error creating actor %s for remote', str(actorClass),
                                exc_info=True)


    def createdRemoteActor(self, remoteActorCreatedMsg, sender):
        lclActorAddr     = remoteActorCreatedMsg.createActorForRemoteMsg.forActor
        lclActorAddrInst = remoteActorCreatedMsg.createActorForRemoteMsg.instanceNum

        self.addressManagers[lclActorAddr.actorAddressString].associateUseableAddress(
            lclActorAddr,
            lclActorAddrInst,
            remoteActorCreatedMsg.createdActorAddress)

        globalName = remoteActorCreatedMsg.createActorForRemoteMsg.globalName
        if globalName:
            self._globalNames[globalName] = remoteActorCreatedMsg.createdActorAddress

        try:
            self.conventionMembers[sender.actorAddressString].createdActor(
                lclActorAddr,
                remoteActorCreatedMsg.createdActorAddress)
        except KeyError:
            # Convention member probably left, no need to record remote created actor anymore
            pass

        if lclActorAddr == self.address:
            # ActorSystem-generated remote actor... nothing more to do (KWQ?)
            self._sendOrQueue(self.systemRmtActorSender[
                remoteActorCreatedMsg.createActorForRemoteMsg.instanceNum],
                              remoteActorCreatedMsg)
        else:
            # Tell the parent Actor about the new child's actual
            # address so that it can update the Actor Address and
            # communicate directly with the new child instead of
            # forwarding messages through the Admin.
            self._sendOrQueue(lclActorAddr, remoteActorCreatedMsg)

        canSend = []
        stillPending = []
        for pendingMsg in self.pendingMessages:
            pmad = pendingMsg[1].toActorAddress.addressDetails
            if isinstance(pmad, ActorLocalAddressExportedToAdmin):
                if lclActorAddr == pmad.creatingActor and lclActorAddrInst == pmad.creationInstNum:
                    canSend.append(pendingMsg)
                else:
                    stillPending.append(pendingMsg)
            else:
                canSend.append(pendingMsg) # strange... expected only PendingActor sends

        self.pendingMessages = stillPending

        for fromActor, msgToSend in canSend:
            self._sendOrQueue(remoteActorCreatedMsg.createdActorAddress, msgToSend.msg, fromActor)


    def failedRemoteActorCreate(self, remoteActorCreateFailedMsg, sender):
        # Remote Actor System (ConventionLeader) could not find a
        # compatible actor system either, so return this to the local
        # caller.
        # KWQ: Negative association for future forwarding?
        cafr = remoteActorCreateFailedMsg.createActorForRemoteMsg
        lclActorAddr = remoteActorCreateFailedMsg.createActorForRemoteMsg.forActor
        if lclActorAddr != self.address:
            self._sendOrQueue(lclActorAddr, remoteActorCreateFailedMsg)
        # KWQ: handle self.pendingMessages (like createdRemoteActor) but dead-letter them instead


    def forwardMessage(self, msgToForward, sender):
        aDetails = msgToForward.toActorAddress.addressDetails
        if isinstance(aDetails, ActorLocalAddressExportedToAdmin):
            faas = aDetails.creatingActor.actorAddressString
            if faas in self.addressManagers:
                toAddr = self.addressManagers[faas].sendingAddress(msgToForward.toActorAddress)
                if toAddr is not None:
                    self._sendOrQueue(toAddr, msgToForward.msg, sender)
                    return
            # If one of the above fails, assume the CreateRemoteActor
            # is not yet completed (or the ForwardMessage got ahead of
            # the PendingActor), so just put this message on the
            # pendingMessages list.
            self.pendingMessages.append((sender, msgToForward))

        else:

            # Strange to be called to forward to a resolveable address, but we can do that.
            self._sendOrQueue(msgToForward.toActorAddress, msgToForward.msg, sender)


    def handleDeadLetter(self, deadLetterAddressMsg, sender):
        lclAddr = deadLetterAddressMsg.deadAddress
        self.logger.info('Admin dead lettering {msg.deadAddress} because: {msg.why}'.format(
            msg=deadLetterAddressMsg))
        if lclAddr == self.deadLetterHandler:
            self.deadLetterHandler = None
        deadSock = addressListener(lclAddr, 'DeadLetterBox')
        self.deadZone.append( (lclAddr, deadSock) )

        for gn in self._globalNames:
            if self._globalNames[gn] == lclAddr:
                del self._globalNames[gn]
                break

        adminCreated = self.addressManagers.get(self.address.actorAddressString, None)
        if adminCreated:
            adminCreated.deadAddress(lclAddr)
        # no response


    def deadLetterRegistration(self, deadLetterRegisterMsg, sender):
        if deadLetterRegisterMsg.enableHandler:
            self.deadLetterHandler = deadLetterRegisterMsg.handlerAddr
        else:
            if deadLetterRegisterMsg.handlerAddr == self.deadLetterHandler:
                self.deadLetterHandler = None


    def registerForConventionNotification(self, notifyConventionMsg, sender):
        if notifyConventionMsg.enableNotification:
            self.conventionNotificationHandler = notifyConventionMsg.handlerAddr

            for each in self.conventionMembers:
                self._sendOrQueue(self.conventionNotificationHandler,
                                  ActorSystemConventionUpdate(
                                      self.conventionMembers[each].remoteAddress,
                                      self.conventionMembers[each].remoteCapabilities,
                                      True))

        else:
            if self.conventionNotificationHandler == notifyConventionMsg.handlerAddr:
                self.conventionNotificationHandler = None


    def handleUnknownWork(self, workMsg, sender):
        self.logger.debug('ignoring unrecognized message' +
                          ' (probably misc response from Actor to System)' +
                          ': {0}'.format(str(workMsg)))


    def shutdown(self):
        # Anything in self.addressManagers that was initially created
        # by "Admin" (actually by the ActorSystem.createActor()) is
        # technically owned by this Admin because it is the only path
        # to communication with that remote Actor, so shutdown those
        # remotes.  Returns True if ready to complete shutdown, or
        # False if shutdown should be delayed.
        shouldWait = False
        adminCreated = self.addressManagers.get(self.address.actorAddressString, None)
        dlh = None
        if adminCreated:
            for rmtAddr in adminCreated.sendingAddresses():
                if rmtAddr == self.deadLetterHandler:
                    # shutdown the dead-letter handler last
                    dlh = rmtAddr
                    continue
                # use raw sends and ignore failures because this will be retried
                sendToAddress(rmtAddr, ActorExitRequest(), self.address)
                shouldWait = True
        # If all other actor shutdowns are complete, go ahead and shutdown the dead-letter handler
        if not shouldWait and dlh:
            sendToAddress(dlh, ActorExitRequest(), self.address)
            shouldWait = True
        if self.conventionAddress:
            if SendStatus.NotSent == sendToAddress(self.conventionAddress,
                                                   ConventionDeRegister(self.address),
                                                   self.address):
                shouldWait = True
        else:
            for each in self.conventionMembers:
                sts = sendToAddress(self.conventionMembers[each].remoteAddress,
                                    ConventionDeRegister(self.address),
                                    self.address)
                if SendStatus.Sent == sts:
                    pass
                    # could del self.conventionMembers[each], but don't modify while under iteration
                else:
                    logging.info('Error saying goodbye to member %s',
                                 self.conventionMembers[each].remoteAddress)
                    # could set shouldWait = True; above would need to remove notified convention members, but
                    # what about members that re-register while we are shutting down
        return not shouldWait


    def handle_deadsockWork(self, work):
        sender, workitem = work
        if self.deadLetterHandler and not isinstance(workitem, (ChildActorExited,
                                                                ActorExitRequest,
                                                                )):
            self._sendOrQueue(self.deadLetterHandler, work)
        # otherwise just discard anything received


    def forwardLogging(self, logmsg, sender):
        self._sendOrQueue(self.loggingActor, logmsg)


    def allowRemoteToExit(self, childexitedmsg, sender):
        pass  # don't need to do anything


    def handle_adminWork(self, work):
        sender, workitem = work

        # Shutdown takes priority
        if isinstance(workitem, SystemShutdown):
            if self.shutdown():
                 # ignore results of send
                sendToAddress(sender, SystemShutdownCompleted(), self.address)
                return True
            notifyAfterTimeout(PENDING_SHUTDOWN_DELAY,
                               self.address,
                               PendingShutdown(PENDING_SHUTDOWN_RETRIES,
                                               sender))
            return False
        if isinstance(workitem, PendingShutdown):
            if self.shutdown() or workitem.cnt == 0:
                # ignore results of send
                sendToAddress(workitem.requester,
                              SystemShutdownCompleted(),
                              self.address)
                return True
            notifyAfterTimeout(PENDING_SHUTDOWN_DELAY,
                               self.address,
                               PendingShutdown(workitem.cnt - 1,
                                               workitem.requester))
            return False

        # Some of the messages that arrive here are Actor responses to
        # ActorSystem().tell() messages; these should just be
        # discarded.  All messages that really need to be handled
        # should be very distinctive.
        self.logger.debug('Admin handling: {0}'.format(workitem.__class__.__name__))
        handler = getattr(self, ([h for h in self._handlers if isinstance(workitem, h[0])] +
                                 [(None, 'handleUnknownWork')])[0][1])
        try:
            handler(workitem, sender)
        except Exception as ex:
            mlog('Admin message %s handling error: %s'%(str(workitem), str(ex)))
            self.logger.error('Admin message %s handling error', str(workitem), exc_info=True)
        return False

    _handlers = [(DeadLetterAddress, 'handleDeadLetter'),
                 (HandleDeadLetters, 'deadLetterRegistration'),
                 (NotifyOnSystemRegistration, 'registerForConventionNotification'),
                 (ConventionRegister, 'handleConventionRegistration'),
                 (ConventionDeRegister, 'handleConventionDeRegistration'),
                 (PendingActor, 'handlePendingActor'),
                 (CreateActorForRemote, 'createActorForRemote'),
                 (RemoteActorCreated, 'createdRemoteActor'),
                 (RemoteActorCreateFailed, 'failedRemoteActorCreate'),
                 (ForwardMessage, 'forwardMessage'),
                 # Primary Actors are not restarted, so allow them to exit
                 (ChildActorExited, 'allowRemoteToExit'),
                 (Thespian_StatusReq, 'getSystemStatus'),
                 (logging.LogRecord, 'forwardLogging'),
             ]

    def checkin_period(self):
        return CONVENTION_RESTART_PERIOD if self._conventionLeaderIsGone else CONVENTION_REREGISTRATION_PERIOD

    def serve_forever(self):
        import select
        retryPeriod = timePeriodSeconds(self.checkin_period())
        try:
            while True:
                if False:
                    if self._queuedsends:
                        if retryPeriod >= timePeriodSeconds(self.checkin_period()):
                            retryPeriod = SEND_QUEUE_RETRY_PERIOD
                        else:
                            # Use of current microsecond adds some
                            # variance so that competing Admins don't
                            # retry in lockstep.
                            retryPeriod = min(retryPeriod * (datetime.datetime.now().microsecond % 10 * 0.1 * RETRY_DELAY_INCREASE),
                                              MAX_RETRY_DELAY)
                    else:
                        retryPeriod = timePeriodSeconds(self.checkin_period())
                else:
                    retryPeriod = 1.23 if self._queuedsends else timePeriodSeconds(self.checkin_period())
                    #retryPeriod = 1.23 if self._queuedsends else CONVENTION_REREGISTRATION_PERIOD.seconds
                (rdy,ig1,ig2) = select.select([self.adminSock] +
                                              [dead[1] for dead in self.deadZone],
                                              [],
                                              [],
                                              retryPeriod)
                for rsock in rdy:
                    retryPeriod = SEND_QUEUE_RETRY_PERIOD
                    try:
                        work = acceptAndReceive(rsock, None)
                        if work:
                            if rsock == self.adminSock:
                                result = self.handle_adminWork(work)
                                if result:
                                    return result
                            else:
                                self.handle_deadsockWork(work)
                    except Exception as ex:
                        mlog('Exception Admining message: %s'%str(ex))
                        self.logger.error('Exception Admining message', exc_info=True)
                # Check on the convention members and re-register if needed
                self._checkConvention()
                # Attempt to send any queued messages
                self._sendQueued()
        except Exception as ex:
            mlog('Exception selecting on Admin activity: %s'%str(ex))
            self.logger.error('Exception selecting on Admin activity', exc_info=True)
        finally:
            self.adminSock.close()

    def getSystemStatus(self, statusReq, sender):
        resp = Thespian_SystemStatus(getattr(self, 'conventionAddress', None),
                                     self.loggingActor,
                                     capabilities = self.system.capabilities)
        for each in self.conventionMembers:
            resp.addConventioneer(self.conventionMembers[each])
        lastreg = getattr(self, '_conventionLastRegisteredAt', None)
        if lastreg:
            resp.setConventionRegisterTime(lastreg)
        for D,S in self.deadZone:
            resp.addDeadLetter(D)
        for F in self.pendingMessages:
            toActor, fwdMsg = F
            resp.addPendingMessage(toActor, fwdMsg.toActorAddress, str(fwdMsg.msg))
        adminCreated = self.addressManagers.get(self.address.actorAddressString, None)
        if adminCreated:
            for rmtAddr in adminCreated.sendingAddresses():
                resp.addChild(rmtAddr)
        self._sendOrQueue(sender, resp)



def startAdmin(adminAddress, adminSock, system, logDefs):
    convAddr = system.capabilities.get('Convention Address.IPv4', None)
    if convAddr:
        convAddr = ActorAddress(SocketActorAddress(socket.AF_INET, socket.SOCK_STREAM,
                                                   socket.IPPROTO_TCP,
                                                   convAddr))
    mlog('startAdmin.1')
    asLogger = startASLogger(logDefs, conventionAddress=convAddr, closeList=[])
    mlog('startAdmin.2')

    # Dirty trick here to completely re-initialize logging in this
    # process... something the standard Python logging interface does
    # not allow via the API.
    oldRoot = logging.root
    logging.root = ThespianLogForwarder(asLogger, adminAddress)
    logging.Logger.root = logging.root
    logging.Logger.manager = logging.Manager(logging.Logger.root)

    logger = logging.getLogger('Thespian.Admin')
    logger.info('ActorSystem Administrator startup @ %s'%str(adminAddress))
    setProcName('admin', adminAddress)

    # n.b. see note regarding similar operation below
    multiprocessing.process._current_process._daemonic = False

    mlog('startAdmin.3')
    admin = ActorSystemAdmin(adminAddress, adminSock, system, asLogger)
    mlog('startAdmin.4')
    admin.serve_forever()
    mlog('startAdmin.5')
    logger.info('ActorSystem Administrator exiting')
    mlog('startAdmin.6')

    # Restore original logging root: do not send any more log messages
    # to the exiting logger

    logging.root = oldRoot
    logging.Logger.root = oldRoot
    logging.Logger.manager = logging.Manager(logging.Logger.root)

    # Shutdown logger; ignore failure
    mlog('startAdmin.7')
    sendToAddress(asLogger, LoggerExitRequest(), adminAddress)
    mlog('startAdmin.8')



# ----------------------------------------------------------------------
# ----------------------------------------------------------------------
# ActorManager
# ----------------------------------------------------------------------
# ----------------------------------------------------------------------


class ActorExited(object): pass

class ActorExitedOnRequest(ActorExited):
    "Actor got an ActorExitRequest"
    def __init__(self, exitRequestedBy):
        self.exitRequestedBy = exitRequestedBy # actor which requested this actor exit
    def __str__(self):
        return 'ActorExitRequest from %s'%str(self.exitRequestedBy)

class ActorExitedException(ActorExited):
    "Actor management got an internal exception and failed... Thespian bug!"
    def __init__(self, exceptionString):
        self.exceptionString = exceptionString
    def __str__(self):
        return 'Actor management exception: "%s"'%str(self.exceptionString)

class ActorExitedBadMessage(ActorExitedException):
    "Actor threw exception while processing this message."
    def __init__(self, msg, sender, exceptionString):
        super(ActorExitedBadMessage, self).__init__(exceptionString)
        self.msg    = msg
        self.sender = sender
    def __str__(self):
        return 'Actor exception handling message "%s" from %s: "%s"'%(str(self.msg),
                                                                      str(self.sender),
                                                                      str(self.exceptionString))

class ActorCreationFailure(ActorExited):
    "Could not instantiate requested Actor"
    def __init__(self, actorClassName):
        self.actorClassName = actorClassName
    def __str__(self):
        return 'Actor class "%s" could not be instantiated'%str(self.actorClassName)


# ......................................................................


class ChildActorInfo:
    def __init__(self, childAddress, targetActorRequirements, childProcess=None):
        self.childAddress            = childAddress
        self.targetActorRequirements = targetActorRequirements
        self.childProcess            = childProcess


class ActorManager(object):
    def __init__(self, instance, myAddress, adminAddress, loggerAddress,
                 systemCapabilities, addressManager):
        self._instance     = instance
        self._address      = myAddress
        self._adminAddress = adminAddress
        self._loggerAddr   = loggerAddress
        self._systemCaps   = systemCapabilities
        self._children     = {} # key=ActorAddressString, val=ChildActorInfo
        self._mySocket     = None
        self._remoteActors = addressManager
        self._msgsPending  = [] # array of (rmtaddr, msg) pending rmtaddr resolution for send
        self._imsgQueue    = [] # array of incoming or to-self (sender, msg) to process
        self._queuedsends = []  # (rmtaddress, msg, fromAddr, timeseq, last_attempt_time)
        self._deadLetterHandler = False
        self._pendingWakeups = {}  # key = datetime for wakeup, value = list of pending wakeupAfter msgs to restart at that time

    @property
    def address(self): return self._address
    @property
    def childAddresses(self): return [c.childAddress for c in self._children.values()]
    @property
    def children(self):
        raise NotImplemented('Cannot get internal child object, try childAddresses instead')
        return self._children
    @children.setter
    def children(self, newChildren):
        raise NotImplemented('Cannot reset children externally')
        self._children = newChildren

    def setInstance(self, newInstance): self._instance = newInstance

    def _sendOne(self, toAddr, msg, fromAddr):
        try:
            sts = sendToAddress(toAddr, msg, fromAddr, 1)  # KWQ config timer not just 1 sec?
            if sts != SendStatus.BadPacket:
                return sts
            # KWQ: log this
            self._imsgQueue.append((toAddr, PoisonMessage(msg)))
            return SendStatus.Sent
        except CannotPickleAddress:
            return SendStatus.NotSent

    def _sendOrQueue(self, rmtAddress, msg, fromActor=None):
        if fromActor is None: fromActor = self.address
        if not self._queuedsends:
            if SendStatus.Sent == self._sendOne(rmtAddress, msg, fromActor):
                return
        self._queuedsends.append( (rmtAddress, msg, fromActor, list(timeseq(300, origDelay=0.47)), datetime.datetime.now() ) )
        # No need to call sendQueued here: will be called from main loop.

    def sendQueued(self):
        failedTo = []
        queued = self._queuedsends
        self._queuedsends = []
        now = datetime.datetime.now()
        for tgt,mesg,frm,delays,lastTried in queued:
            if tgt not in failedTo:
                elapsed = (((now - lastTried) * 100).seconds * 1.0) / 100.0
                if elapsed >= delays[0]:
                    if SendStatus.Sent == self._sendOne(tgt, mesg, frm):
                        continue
                    delays = delays[1:]
                    lastTried = now
                failedTo.append(tgt)
            if delays:
                self._queuedsends.append( (tgt,mesg,frm,delays,lastTried) )
            else:
                logging.getLogger('Thespian.Actor').warning('Dropped send of %s from %s to %s; could not send in a reasonable time period.',
                                    str(mesg), str(frm), str(tgt))

        return 0 != len(self._queuedsends)


    def createActor(self, newActorClass, targetActorRequirements, globalName):
        if not globalName:
            naa,nas = newActorAddress(name = '%s Actor'%newActorClass.__name__)
            try:
                na = _createActor(newActorClass, targetActorRequirements,
                                  naa, nas, self.address,
                                  self._adminAddress, self._loggerAddr,
                                  closeList=[self._mySocket],
                                  systemCapabilities=self._systemCaps)
                self._remoteActors.registerUseableAddress(naa)
                time.sleep(CREATE_ACTOR_THROTTLE_DELAY)
                self._children[naa.actorAddressString] = ChildActorInfo(naa, targetActorRequirements, na)
                return naa

            except NoCompatibleSystemForActor:
                pass

        naa = self._remoteActors.createLocalAddress()
        na = None
        self._sendOrQueue(self._adminAddress,
                          PendingActor(newActorClass, self.address,
                                       naa.addressDetails.addressInstanceNum,
                                       targetActorRequirements, globalName))
        self._children[naa.actorAddressString] = ChildActorInfo(naa, targetActorRequirements, na)
        return naa

    def actor_send(self, toActorAddress, msg):
        rmtAddr = self._remoteActors.sendingAddress(toActorAddress)
        try:
            if rmtAddr:
                if rmtAddr == self.address:
                    self._imsgQueue.append((self.address, msg))
                else:
                    self._sendOrQueue(rmtAddr, msg)
            else:
                self._sendOrQueue(self._adminAddress,
                                  ForwardMessage(
                                      self._remoteActors.exportAddress(toActorAddress),
                                      msg))
        except CannotPickleAddress:
            # msg contains an address that cannot be pickled yet.
            # Admin cannot forward for use because we can't pickle
            # the address(es) in the message to send to the Admin,
            # so it must be saved locally.
            self._msgsPending.append( (toActorAddress, msg) )

    def handle_message(self, msg, sender):
        # Returns:
        #   None - message handled successfully
        #   ActorExited instance
        if isinstance(msg, RemoteActorCreated):
            self._remoteActors.associateUseableAddress(msg.createActorForRemoteMsg.forActor,
                                                       msg.createActorForRemoteMsg.instanceNum,
                                                       msg.createdActorAddress)
            # Now that a new translation is known, retry sending any messages that were
            # previously unpickleable.
            pendings = self._msgsPending
            self._msgsPending = []
            for addr, msg in pendings:
                self.actor_send(addr, msg)
            return None
        elif isinstance(msg, RemoteActorCreateFailed):
            failedActorAddr = self._remoteActors.getLocalAddress(msg.createActorForRemoteMsg.instanceNum)
            # Let Actor know that remote actor creation failed.  Set
            # sender address to ourselves.
            self._imsgQueue.append((self.address, ChildActorExited(failedActorAddr)))
            return None
        elif isinstance(msg, Thespian_StatusReq):
            self.getActorStatus(msg, sender)
            return None
        self._remoteActors.registerUseableAddress(sender)
        try:
            self._instance.receiveMessage(msg, sender)
        except Exception as ex:
            logger = logging.getLogger('Thespian.Actor')
            import traceback
            mlog('Actor %s message handling exception: %s'%(str(self._instance), traceback.format_exc()))
            logger.warning('Actor %s message handling exception', str(self._instance), exc_info=True)
            return ActorExitedBadMessage(msg, sender, str(ex))
        if isinstance(msg, ChildActorExited):
            if msg.childAddress.actorAddressString in self._children:
                del self._children[msg.childAddress.actorAddressString]
            else:
                pass
                # logger = logging.getLogger('Thespian.Actor')
                # logger.warning('Warning, %s not in list of known children (%s)'%(msg.childAddress.actorAddressString,
                #                                                                  self._children.keys()))
        return None

    def logger(self, name=None): return logging.getLogger(name)


    def wakeupAfter(self, timePeriod):
        wakeupTime = datetime.datetime.now() + timePeriod
        self._pendingWakeups.setdefault(wakeupTime, []).append( (self.address, WakeupMessage(timePeriod)) )


    def _realizeWakeups(self):
        "Find any expired wakeups and queue them to the send processing queue"
        now = datetime.datetime.now()
        removals = []
        for wakeupTime in self._pendingWakeups:
            if wakeupTime > now:
                continue
            self._imsgQueue.extend(self._pendingWakeups[wakeupTime])
            removals.append(wakeupTime)
        for each in removals:
            del self._pendingWakeups[each]


    def serve_forever(self):
        # This is the main loop of an Actor.  It waits for a received
        # message and then calls the Actor to handle that message.  If
        # the message is special (e.g. ActorExitRequest it will
        # perform the associated operations.
        #
        # Returns an instance of ActorExited.

        #    ActorAddress - ActorExitRequest received from this address
        #    string - exiting on exception in management, string is exception message
        #    (msg, sender, string) - Actor exception handling message from sender
        try:
            with addressListener(self._address,
                                 '%s Actor'%self._instance.__class__.__name__) as aSocket:
                self._mySocket = aSocket
                while True:
                    # first send queued
                    moreToSend = self.sendQueued()
                    # then check wakeups
                    self._realizeWakeups()
                    # and then handle the incoming message queue
                    if not self._imsgQueue:
                        # finally listen, unless there's still more to send, like actor restart/busy
                        now = datetime.datetime.now()
                        msg = acceptAndReceive(aSocket,
                                               min([timePeriodSeconds(now, P) for P in self._pendingWakeups] +
                                                   [SEND_QUEUE_RETRY_PERIOD if moreToSend else None]))
                        self._realizeWakeups()  # Include any wakeups that occurred while waiting for this message
                        if msg:
                            self._imsgQueue.append(msg)
                    iQ = self._imsgQueue
                    self._imsgQueue = []
                    while iQ:
                        sender, fullMsg = iQ[0]
                        del iQ[0]
                        mlog('Delivering %s from %s'%(str(fullMsg), str(sender)))
                        sts = self.handle_message(fullMsg, sender)
                        if isinstance(fullMsg, ActorExitRequest):
                            if self._deadLetterHandler:
                                # De-register as a dead letter handler
                                self._sendOrQueue(self._adminAddress,
                                                    HandleDeadLetters(self.address, False))
                            return ActorExitedOnRequest(sender)
                        elif sts is not None:
                            return sts
                    # Check for children that may have died without being able to notify us.
                    currentChildren = self._children.keys()  # because we can't change the dict if we are iterating over it
                    for child in currentChildren:
                        if child not in self._children: continue
                        childProc = self._children[child].childProcess
                        if childProc is not None and not childProc.is_alive():
                            childProc.join(0)
                            self._children[child].childProcess = None
                            # KWQ: what if child wanted to exit...
                            cae = ChildActorExited(self._children[child].childAddress)
                            sts = self.handle_message(cae, self._children[child].childAddress)
                            if sts is not None:
                                return sts
        except Exception as ex:
            logger = logging.getLogger('Thespian.Actor')
            mlog('Actor %s management exiting on exception: %s'%(str(self._instance), str(ex)))
            logger.warning('Actor %s management exiting on exception', str(self._instance), exc_info=True)
            return ActorExitedException(str(ex))
        finally:
            self._mySocket = None


    def getActorStatus(self, statusReq, sender):
        resp = Thespian_ActorStatus(self.address,
                                    self._instance.__class__.__name__,
                                    self._adminAddress,
                                    pendingWakeups = self._pendingWakeups)
        for C in self._children:
            resp.addChild(self._children[C].childAddress,
                          -1 if self._children[C].childProcess is None else self._children[C].childProcess.pid)
        for F in self._msgsPending:
            resp.addPendingMessage(F[0], str(F[1]))
        self._sendOrQueue(sender, resp)

    def handleDeadLetters(self, addr, enable):
        self._deadletterhandler = enable
        self._sendOrQueue(self._adminAddress, HandleDeadLetters(addr, enable))

    def notifyOnSystemRegistrationChanges(self, addr, enable):
        self._sendOrQueue(self._adminAddress,
                          NotifyOnSystemRegistration(addr, enable))


def _instantiateActor(mgr, actorClass, logger):
    try:
        try:
            if type(actorClass) == type(""):
                actorClass = eval(actorClass)
            actor = actorClass(childActors=(None if not mgr.childAddresses else mgr.childAddresses))
        except TypeError as te:
            if "unexpected keyword argument 'childActors'" in str(te):
                # Some actors do not take a childActors keyword argument, so init without it
                actor = actorClass()
            else:
                logger.warning("Got a TypeError attempting to instantiate Actor %s",
                               str(actorClass), exc_info=True)
                return None
    except Exception as ex:
        mlog("Unable to instantiate Actor %s: %s"%(str(actorClass), str(ex)))
        logger.warning("Unable to instantiate Actor %s", str(actorClass), exc_info=True)
        return None
    mgr.setInstance(actor)
    actor._myRef = mgr
    return actor


def startupActor(actorClass, actorAddress, parentAddress,
                 adminAddress, loggerAddress,
                 closeList, systemCapabilities):
    # Dirty trick here to completely re-initialize logging in this
    # process... something the standard Python logging interface does
    # not allow via the API.  We also do not want to run
    # logging.shutdown() because (a) that does not do enough to reset,
    # and (b) it shuts down handlers, but we want to leave the parent's
    # handlers alone.
    logging.root = ThespianLogForwarder(loggerAddress, actorAddress)
    logging.Logger.root = logging.root
    logging.Logger.manager = logging.Manager(logging.Logger.root)

    logger = logging.getLogger('Thespian.Manager')

    # Dirty trick here to workaround multiprocessing trying to impose
    # an unnecessary restriction.  A process should be set daemonic
    # before start() if the parent shouldn't track it (an specifically
    # automatically join() the subprocess on exit).  For Actors, the
    # parent exists independently of the child and the ActorSystem
    # manages them, so daemonic processes are desired.  However,
    # multiprocessing imposes a restriction that daemonic processes
    # cannot create more processes.  The following reaches deep into
    # the implementation of the multiprocessing module to override
    # that restriction.  This process was already started as daemonic,
    # and it's detached from its parent.  The following simply clears
    # that flag locally so that other processes can be created from
    # this one.
    multiprocessing.process._current_process._daemonic = False

    for H in closeList:
        if H:
            try:
                H.close()
            except:
                pass

    logger.debug('Starting Actor %s @ %s with admin @ %s'%(str(actorClass),
                                                           str(actorAddress), str(adminAddress)))
    setProcName(actorClass.__name__, actorAddress)
    servedResult = None
    addressManager = ActorAddressManager(actorAddress)
    mgr = ActorManager(None, actorAddress, adminAddress, loggerAddress,
                       systemCapabilities, addressManager)

    actor = _instantiateActor(mgr, actorClass, logger)

    while actorClass:

        if actor:

            if isinstance(servedResult, ActorExitedBadMessage):
                # Previously exited while handling a message, retry
                # handling that message.  A second failure means send
                # that message back as a PoisonMessage.
                servedResult = mgr.handle_message(servedResult.msg, servedResult.sender)
                if isinstance(servedResult, ActorExitedBadMessage):
                    mlog('Got Message Handling Actor Exit, send back Poison message')
                    for delay in timeseq(120):
                        if sendToAddress(servedResult.sender,
                                         PoisonMessage(servedResult.msg),
                                         actorAddress) in [SendStatus.Sent,
                                                           SendStatus.BadPacket]:
                            break
                        time.sleep(delay)
                    servedResult = None
                    # # do not serve_forever()... this Actor instance
                    # # just died, but also do not retry this message
                    # # any more.
                    # servedResult = ActorExitedException('Bad message tried twice'
                    #                                     ', returned to sender'
                    #                                     ' in PoisonMessage wrapper.')
            else:
                servedResult = None

            if servedResult is None:

                # No previous exit, or else the previous exited
                # message was handled successfully, so enter the main
                # serve loop for this Actor instance.

                servedResult = mgr.serve_forever()  # <--------- gold!!

                if not isinstance(servedResult, ActorExited):
                    logger.warning('Unrecognized servedResult: %s', str(servedResult))
                    servedResult = ActorExitedException("unexpected exit from Actor management")

        else:
            servedResult = ActorCreationFailure(str(actorClass))

        if True: #parentAddress:
            if isinstance(servedResult, ActorExitedOnRequest) and \
               servedResult.exitRequestedBy == parentAddress:
                # Parent asked us to exit, so no need to inform parent
                # of exit via ChildActorExited, just leave quickly.
                logger.info('Actor %s @ %s exited on request from parent (%s)'%(str(actorClass),
                                                                                str(actorAddress),
                                                                                str(parentAddress)))
                actorClass = None
            elif not isinstance(servedResult, ActorExitedBadMessage):
                logger.info('Actor %s @ %s exited, notifying parent @ %s (%s)'%(str(actorClass),
                                                                                str(actorAddress),
                                                                                str(parentAddress),
                                                                                str(servedResult)))

                if parentAddress:
                    for delay in timeseq(180):
                        if sendToAddress(parentAddress,
                                         ChildActorExited(actorAddress),
                                         actorAddress) in [SendStatus.Sent, SendStatus.BadPacket]:
                            break
                        time.sleep(delay)
                actorClass = None

        else:
            if not isinstance(servedResult, ActorExitedBadMessage):
                logger.info('Primary Actor %s @ %s exited (%s).'%(str(actorClass),
                                                                  str(actorAddress),
                                                                  str(servedResult)))
                actorClass = None

    logger.info('Actor @ %s exiting without replacement'%str(actorAddress))
    for delay in timeseq(180):
        if sendToAddress(adminAddress,
                         DeadLetterAddress(actorAddress, str(servedResult)),
                         adminAddress) in [SendStatus.Sent,
                                           SendStatus.BadPacket]:
            break
        time.sleep(delay)

    for cas in mgr.childAddresses:
        childAddr = addressManager.sendingAddress(cas)
        if childAddr:
            if sendToAddress(childAddr, ActorExitRequest(), adminAddress) in [SendStatus.Sent,
                                                                              SendStatus.BadPacket]:
                continue
        for delay in timeseq(180):
            if sendToAddress(adminAddress,
                             ForwardMessage(addressManager.exportAddress(cas),
                                            ActorExitRequest()),
                             adminAddress) in [SendStatus.Sent,
                                               SendStatus.BadPacket]:
                break


def _createActor(actorClass, targetActorRequirements,
                 newActorAddr, newActorSocket, parentAddr,
                 adminAddr, loggingAddr, closeList=[],
                 systemCapabilities=None):
    if newActorSocket: newActorSocket.close()  # created actor process will re-open at same address

    if hasattr(actorClass, "actorSystemCapabilityCheck"):
        if not actorClass.actorSystemCapabilityCheck(systemCapabilities or {},
                                                     targetActorRequirements):
            raise NoCompatibleSystemForActor(actorClass, "no system has compatible capabilities")

    # n.b. see note regarding similar operation below
    multiprocessing.process._current_process._daemonic = False
    na = multiprocessing.Process(target=startupActor,
                                 args=(actorClass,
                                       newActorAddr, parentAddr,
                                       adminAddr, loggingAddr,
                                       closeList, systemCapabilities))
    na.daemon = True
    attempts = 0
    while True:
        attempts = attempts + 1
        try:
            na.start()
            # Also note that while non-daemonic children cause the
            # current process to automatically join() those children
            # on exit, daemonic children are sent a terminate()
            # operation (usually indicated by a SIGTERM under unix or
            # TERMINATE indicator under windows.  To avoid this, use
            # another dirty trick and remove all children from the
            # _current_process._children list.
            multiprocessing.process._current_process._children.remove(na)
            break
        except OSError as ex:
            if ex.errno == errno.EAGAIN:
                if attempts < MAX_ACTOR_STARTUP_RETRIES:
                    time.sleep(1)
                    continue
            raise
    #logging.getLogger('Thespian').debug('New Actor Address: %s for Actor %s'%(str(newActorAddr), str(actorClass)))
    return na


# ----------------------------------------------------------------------


class ActorSystemBase:

    """This is a singleton, and it must manage all the processes in this
       ActorSystem, so it will create a separate process to actually
       run this base.  Although this object is a "singleton", the
       splitting of this process into multiple processes to represent
       other actors will result in an instance of this object in each
       process (although it will only be initialized once).  The
       separate ActorSystemAdmin process will be the actual singleton
       in this environment.

       All ActorAddresses generated via newActor and newPrimaryActor
       are local to this ActorSystemBase instance.  Any and *all*
       messages sent to other Actors must be pickled and unpickled;
       this allows the pickling/unpickling process to translate an
       ActorAddress from a local representation to a global or remote
       representation.

       Specific capabilities recognized by this system:

          "Admin Port"

              Specifies the network port for the ActorSystem
              administration functionality.  The default value is
              14334.  If multiple ActorSystems communicate with each
              other, they will contact each other on this port.

          "Convention Address.IPv4"

              Specifies the IPv4 network address of a remote
              ActorSystem that this ActorSystem should register with.
              The remote address should be the Convention Leader.  All
              ActorSystems within the Convention are aware of each
              other and the other's capabilities; createActor
              operations can cause Actors to be created in remote
              ActorSystems that are part of the Convention.

    """

    def __init__(self, system, logDefs = None):
        self._primaryActors = {}   # key=actorAddressString, value=(ActorAddress, Process or None)
        self.capabilities = system.capabilities

        try:
            adminAddr, adminSocket = newActorAddress(system.capabilities.get('Admin Port', 14334),
                                                     name='ActorAdmin')

            # n.b. see note regarding similar operation below
            multiprocessing.process._current_process._daemonic = False
            admin = multiprocessing.Process(target=startAdmin,
                                            args=(adminAddr, adminSocket, system, logDefs))
            admin.daemon = True
            #hack to allow web server to return response before child processes complete
            multiprocessing.process._current_process._daemonic = False
            admin.start()
            # admin must be explicitly shutdown and is not automatically
            # started when this current process exits.  See similar note above.
            multiprocessing.process._current_process._children.remove(admin)
            adminSocket.close()  # still open in started Process
        except InvalidActorAddress as ex:
            adminAddr = ex.actorAddress
            logging.info('Using running ActorSystem at %s', adminAddr)

        try:
            rsp = askToAddress(adminAddr, Thespian_StatusReq(), 5)
        except SendStatus.BadPacket:
            raise InvalidActorAddress(adminAddr,
                                      'incompatible Thespian version running at %s'%str(adminAddr))
        if not isinstance(rsp, Thespian_SystemStatus):
            raise InvalidActorAddress(adminAddr,
                                      'not an ActorSystem? (got %s)'%str(rsp))

        self.adminAddr = adminAddr
        self.loggingAddr = rsp.loggingActor
        self._primaryAddresses = ActorAddressManager(self.adminAddr)

    def __getstate__(self):
        raise CannotPickle('ActorSystem cannot be Pickled.')

    def _sendToAdmin(self, msg):
        sendOrFail(self.adminAddr, msg, self.adminAddr, 60)

    def shutdown(self):
        logger = logging.getLogger('Thespian')
        logger.warning('ActorSystem (multiprocess) shutdown initiated.')
        sent_shutdown_reqs = False
        for cas in self._primaryAddresses.sendingAddresses():
            try:
                sendToAddress(cas, ActorExitRequest(), self.adminAddr, 6)
                sent_shutdown_reqs = True
            except:
                pass
        # Admin shutdown will send exit requests to anything _primaryAddresses couldn't supply
        # Not all primary actors are registered in primaryAddresses though
        for cas in self._primaryActors:
            paa = self._primaryAddresses.sendingAddress(self._primaryActors[cas][0])
            if paa:
                try:
                    # Again, ignore errors in shutdown process
                    sendToAddress(paa, ActorExitRequest(), self.adminAddr, 6)
                    sent_shutdown_reqs = True
                except:
                    pass
        # Now shutdown the Admin itself and exit everything.
        if sent_shutdown_reqs:
            time.sleep(PENDING_SHUTDOWN_DELAY)  # allow children time to shutdown
        # n.b. if askToAddress raises SendStatus.BadPacketError, let that propagate
        if askToAddress(self.adminAddr, SystemShutdown(),
                        (PENDING_SHUTDOWN_RETRIES *
                         (PENDING_SHUTDOWN_DELAY +
                          (SEND_CONNREFUSED_RETRIES * SEND_CONNREFUSED_BACKOFF_PERIOD * 5)
                         ))) is None:
            raise RuntimeError('Unable to shutdown multiprocess Actor System Admin.')
        # for cas in self._primaryActors:
        #     if self._primaryActors[cas][1]:
        #         self._primaryActors[cas][1].join()

        # Do not join(); it's a daemon and might already be gone, in which case this hangs.
        #self.admin.join()

        logger.info('Admin exited, ActorSystem shutdown complete.')

    def newPrimaryActor(self, actorClass, targetActorRequirements, globalName):

        if not globalName:
            # n.b. see note regarding similar operation below
            multiprocessing.process._current_process._daemonic = False
            naa,nas = newActorAddress(name='Primary %s Actor'%(actorClass.__name__))

            try:
                na = _createActor(actorClass, targetActorRequirements,
                                  naa, nas, None, self.adminAddr, self.loggingAddr,
                                  systemCapabilities=self.capabilities)
                self._primaryAddresses.registerUseableAddress(naa)
                time.sleep(CREATE_ACTOR_THROTTLE_DELAY)
                self._primaryActors[naa.actorAddressString] = (naa, na)
                return naa
            except NoCompatibleSystemForActor:
                pass
            except Exception as ex:
                logger = logging.getLogger('Thespian')
                mlog('ERROR creating Primary Actor %s: %s'%(str(actorClass), str(ex)))
                logger.error('ERROR creating Primary Actor %s', str(actorClass), exc_info=True)
                #KWQ: dead letter the return address?  Right now, just fall through and see if Admin can do it.

        naa = self._primaryAddresses.createLocalAddress()
        na = None
        try:
            rmtActorCreated = askToAddress(self.adminAddr,
                                           PendingActor(actorClass, self.adminAddr,
                                                        naa.addressDetails.addressInstanceNum,
                                                        targetActorRequirements, globalName))
        except SendStatus.BadPacketError:
            raise NoCompatibleSystemForActor(actorClass,
                                             'Incompatible with admin for remote actor creation')
        if isinstance(rmtActorCreated, RemoteActorCreated):
            naa = rmtActorCreated.createdActorAddress
        else:  # should be a RemoteActorCreateFailed
            raise NoCompatibleSystemForActor(actorClass,
                                             'No RemoteActorCreated: %s (%s)'%(
                                                 str(rmtActorCreated),
                                                 type(rmtActorCreated)))
        self._primaryActors[naa.actorAddressString] = (naa, na)
        return naa

    def tell(self, anActor, msg):
        rmtAddr = self._primaryAddresses.sendingAddress(anActor)
        try:
            if rmtAddr:
                sendOrFail(rmtAddr, msg, self.adminAddr, 180)
            else:
                sendOrFail(self.adminAddr,
                           ForwardMessage(self._primaryAddresses.exportAddress(anActor), msg),
                           self.adminAddr, 60)
        except Exception as ex:
            logging.getLogger('Thespian').error('Tell %s to %s failed: %s'%(str(msg), str(anActor), str(ex)))
            mlog('Tell %s to %s failed: %s'%(str(msg), str(anActor), str(ex)))
            raise

    def ask(self, anActor, msg, timeout):
        rmtAddr = self._primaryAddresses.sendingAddress(anActor)
        # n.b. if askToAddress raises SendStatus.BadPacketError, let that propagate
        try:
            if rmtAddr:
                return askToAddress(rmtAddr, msg, timeout)
            else:
                return askToAddress(self.adminAddr,
                                    ForwardMessage(self._primaryAddresses.exportAddress(anActor), msg),
                                    timeout)
        except Exception as ex:
            logging.getLogger('Thespian').error('Ask %s of %s failed: %s'%(str(msg), str(anActor), str(ex)))
            mlog('Ask %s of %s failed: %s'%(str(msg), str(anActor), str(ex)))
            raise

    def dupLogToFile(self, fname):
        fdup = LoggerFileDup
        fdup.fname = fname
        logging.getLogger('Thespian').critical(fdup)
        return fname
