import time
import logging
from datetime import datetime, timedelta
from thespian.actors import *
from thespian.system.systemBase import thesplog
from thespian.system.systemCommon import systemCommonBase
from thespian.system.messages.status import Thespian_SystemStatus
from thespian.system.messages.admin import *
from thespian.system.transport import *


class AdminCore(systemCommonBase):

    def __init__(self, preppedEndpoint, transportClass, address, capabilities, logdefs):
        thesplog('starting Admin', level=logging.DEBUG)
        super(AdminCore, self).__init__(address, transportClass, preppedEndpoint)
        self.capabilities = capabilities
        self.logdefs      = logdefs
        self._pendingChildren = {}  # key = childLocalAddr instance #, value = PendingActorEnvelope
        # Things that help us look like an Actor, even though we're not
        self._sourceHash  = None
        thesplog('Admin started @ %s', self.transport.myAddress, level=logging.INFO)


    def run(self):
        try:
            self.transport.run(self.handleIncoming, None) #timedelta(seconds=10))
        except Exception as ex:
            import traceback
            thesplog('ActorAdmin uncaught exception: %s', traceback.format_exc(),
                     level=logging.ERROR, exc_info=True)
        thesplog('Admin time to die', level=logging.DEBUG)


    def handleIncoming(self, envelope):
        thesplog('Admin of %s for %s', envelope.message, envelope.sender, level=logging.DEBUG)
        return getattr(self,
                       'h_' + envelope.message.__class__.__name__, self.unrecognized)(envelope)


    def unrecognized(self, envelope):
        thesplog("Admin got incoming %s from %s; discarded because I don't know how to handle it!",
                 envelope.message, envelope.sender, level=logging.WARNING)
        return True


    def h_Thespian_StatusReq(self, envelope):
        resp = Thespian_SystemStatus(None, # KWQ: conventionAddr
                                     None, # KWQ: self.loggingActor address?
                                     capabilities = self.capabilities)
        # KWQ: add conventioneers
        # KWQ: resp.setConventionRegisterTime from last time
        # KWQ: resp.addDeadLetter for each
        # KWQ: resp.addPendingMessage(toActor, fwdMsg.toActorAddress, str(fwdMsg.msg)) for each in pendingMessages
        # KWQ: if self.addressManagers.get(self.myAddress.actorAddressString, None):
        #          for rmtAddr in ^^.sendingAddresses():
        #              resp.addChild(rmtAddr)
        self._send_intent(TransmitIntent(envelope.sender, resp,
                                         onSuccess=self._checkNextTransmit,
                                         onError = self._checkNextTransmit))
        return True


    def h_SystemShutdown(self, envelope):
        self._exiting = envelope.sender

        # Send failure notices and clear out any pending children.  If
        # any pending child ready notifications are received after
        # this, they will automatically be sent an ActorExitRequest.
        for each in self._pendingChildren:
            thesplog('Cancelling pending child to %s', self._pendingChildren[each].sender)
            self._send_intent(
                TransmitIntent(self._pendingChildren[each].sender,
                               PendingActorResponse(each.message.forActor,
                                                    each.message.instanceNum,
                                                    isSuccess = False),
                               onSuccess=self._checkNextTransmit,
                               onError = self._checkNextTransmit))
        self._pendingChildren = []

        children = list(self._addrManager.sendingAddresses())

        if not children:
            self._sayGoodbye()
            self.transport.abort_run(drain=True)
            return True

        # Now shutdown any direct children
        for each in children:
            thesplog('Asking child @ %s to shutdown', each)
            self._send_intent(
                TransmitIntent(each, ActorExitRequest(recursive=True),
                               onSuccess=self._checkNextTransmit,
                               onError = self._checkNextTransmit))

        # Once children confirm their exits the callback will shutdown the Admin.
        thesplog('Admin pending shutdown on children shutdown confirmations', level=logging.DEBUG)
        return True


    def _sayGoodbye(self):
        self._send_intent(TransmitIntent(self._exiting, SystemShutdownCompleted(),
                                         onSuccess=self._checkNextTransmit,
                                         onError = self._checkNextTransmit))


    def h_ChildActorExited(self, envelope):
        return self._handleChildExited(envelope.message.childAddress)


    def h_PendingActor(self, envelope):

        # Note, both Admin and remote requester will have a local
        # child address for the child (with a different instanceNumber
        # each).
        childAddr = self._addrManager.createLocalAddress()

        self._startChildActor(childAddr, envelope.message.actorClass,
                              # KWQ: childRequirements?
                              sourceHash = envelope.message.sourceHash)
        self._pendingChildren[childAddr.addressDetails.addressInstanceNum] = envelope
        # transport will contrive to call _pendingActorReady when the
        # child is initialized and connected to this parent.
        return True


    def _pendingActorReady(self, childInstance, actualAddress):
        thesplog('Admin pending actor ready %s is %s', childInstance, actualAddress)
        if childInstance not in self._pendingChildren:
            thesplog('Pending actor is ready at %s for %s but latter is unknown'
                     '; sending child a shutdown',
                     actualAddress, childInstance, level=logging.WARNING)
            self._send_intent(
                TransmitIntent(actualAddress, ActorExitRequest(recursive=True),
                               onSuccess=self._checkNextTransmit,
                               onError = self._checkNextTransmit))
            return

        requestEnvelope = self._pendingChildren[childInstance]
        del self._pendingChildren[childInstance]
        self._addrManager.registerUseableAddress(actualAddress)
        self._addrManager.associateUseableAddress(self.myAddress, childInstance, actualAddress)
        # n.b. childInstance is for this Admin, but caller's childInstance is in original request
        thesplog('Admin pending actor ready %s is %s telling %s (%s #%s)', childInstance, actualAddress, requestEnvelope.sender, requestEnvelope.message.forActor, requestEnvelope.message.instanceNum)
        self._send_intent(
            TransmitIntent(requestEnvelope.sender,
                           PendingActorResponse(requestEnvelope.message.forActor,
                                                requestEnvelope.message.instanceNum,
                                                isSuccess = True,
                                                actualAddress = actualAddress),
                           onSuccess=self._checkNextTransmit,
                           onError = self._checkNextTransmit))



"""
        thesplog('ActorSystem newPrimaryActor(%s, globalName %s) TBD',
                 actorClass, globalName, level=logging.WARNING)
        raise NoCompatibleSystemForActor(actorClass,
                                         'Actor %s has requirements: %s'%(
                                             str(actorClass),
                                             str(targetActorRequirements)))
"""

