'''The systemBase provides the base class implementation for standard
system Base implementations.  This systemBase itself is not intended
to be instantiated as the regular Thespian System Base, but instead it
provides a base class that should be subclassed by the various System
Base implementations.

'''

import logging, logging.handlers
try:
    from logging.config import dictConfig
except ImportError:
    # Old python that doesn't contain this...
    from thespian.system.dictconfig import dictConfig

from thespian.actors import *
from thespian.system import *
from thespian.system.messages import TellMessage
from thespian.system.messages.admin import *
from thespian.system.transport import *

from datetime import datetime, timedelta
import os

MAX_SYSTEM_SHUTDOWN_DELAY    = timedelta(seconds=50)
MAX_CHILD_ACTOR_CREATE_DELAY = timedelta(seconds=5)
MAX_ADMIN_STATUS_REQ_DELAY   = timedelta(seconds=2)


def thesplog(msg, *args, **kw):
    levelstr = lambda l: { logging.DEBUG: 'dbg',
                           logging.INFO:  'I',
                           logging.WARNING: 'Warn',
                           logging.ERROR:   'ERR',
                           logging.CRITICAL: 'CRIT' }.get(l, '??')
    with open('/tmp/Thespian.log','a') as lf:
        lf.write('%s p%s %-4s %s\n'%(str(datetime.now()), os.getpid(),
                                      levelstr(kw.get('level', logging.INFO)), str(msg%args)))
    logging.getLogger('Thespian.System').log(kw.get('level', logging.INFO), msg, *args,
                                             exc_info=kw.get('exc_info', False),
                                             extra=kw.get('extra', None))


class systemBase(object):

    """This is the systemBase base class that various Thespian System Base
       implementations should subclass.  The System Base is
       instantiated by each process that wishes to utilize an Actor
       System and runs in the context of that process (as opposed to
       the System Admin that may run in its own process).

       Depending on the System Base implementation chosen by that
       process, the instantiation may be private to that process or
       shared by other processes; in the former case, there will be an
       instance of this class in each process accessing the shared
       Actor System, representing the Portal between the "external"
       environment of that process and the shared Actor System
       Implementation.

       All ActorAddresses generated via newActor and newPrimaryActor
       are local to this ActorSystemBase instance.  Any and *all*
       messages sent to other Actors must be able to be appropriately
       serialized; this allows the pickling/unpickling process to
       translate an ActorAddress from a local representation to a
       global or remote representation.

    """

    def __init__(self, system, logDefs = None):
        # Expects self.transport has already been set by subclass __init__
        self.adminAddr = self.transport.getAdminAddr(system.capabilities)
        self._startAdmin(self.adminAddr, self.transport.myAddress, system.capabilities, logDefs)
        self._numPrimaries = 0
        if not self._verifyAdminRunning():
            # KWQ: import this
            raise InvalidActorAddress(self.adminAddr,
                                      'not a valid or useable ActorSystem Admin')
            # KWQ: more details? couldn't start @ addr? response was ? instead of expected Thespian_SystemStatus?


    def _verifyAdminRunning(self):
        """Returns boolean verification that the Admin is running and
           available.  Will query the admin for a positive response,
           blocking until one is received.
        """
        self.transport.scheduleTransmit(
            TransmitIntent(self.adminAddr, Thespian_StatusReq(), onError=self._verifySendFailed))
        response = self.transport.run(None, MAX_ADMIN_STATUS_REQ_DELAY)
        return not getattr(self, '_VERIFYFAILED', False) and \
            response and \
            response.sender == self.adminAddr and \
            isinstance(response.message, Thespian_SystemStatus)

    def _verifySendFailed(self, msg):
        self._VERIFYFAILED = True
        self.transport.abort_run()



    def __getstate__(self):
        raise CannotPickle('ActorSystem cannot be Pickled.')

    def shutdown(self):
        thesplog('ActorSystem shutdown initiated.', level=logging.INFO)
        time_to_quit = datetime.now() + MAX_SYSTEM_SHUTDOWN_DELAY
        self.transport.scheduleTransmit(
            TransmitIntent(self.adminAddr, SystemShutdown(), onError=self._shutdownSendFailed))
        while datetime.now() < time_to_quit:
            response = self.transport.run(None, time_to_quit - datetime.now())
            if getattr(self, '_TASF', False):
                thesplog('Could not send shutdown request to Admin'
                         '; aborting but not necessarily stopped',
                         level=logging.WARNING)
                return
            if response:
                if isinstance(response.message, SystemShutdownCompleted):
                    break
                else:
                    thesplog('Expected shutdown completed message, got: %s', response.message,
                             level=logging.WARNING)
            else:
                thesplog('No response to Admin shutdown request; Actor system not completely shutdown',
                         level=logging.WARNING)
        thesplog('ActorSystem shutdown complete.')

    def _shutdownSendFailed(self, msg):
        self._TASF = True
        self.transport.abort_run()


    def newPrimaryActor(self, actorClass, targetActorRequirements, globalName, sourceHash=None):
        self._numPrimaries = self._numPrimaries + 1
        self._pcrFAILED = False
        self.transport.scheduleTransmit(
            TransmitIntent(self.adminAddr,
                           PendingActor(actorClass,
                                        None, self._numPrimaries,
                                        targetActorRequirements,
                                        globalName=globalName,
                                        sourceHash=sourceHash),
                           onError = self._newPrimarySendFailed))
        response = self.transport.run(None, MAX_CHILD_ACTOR_CREATE_DELAY)
        if self._pcrFAILED:
            raise ActorSystemFailure("Could not request new Actor from Admin")
        if response:
            if isinstance(response.message, PendingActorResponse):
                if response.message.isSuccess:
                    thesplog('PrimaryActor %s @ %s', actorClass, response.message.actualAddress)
                    return response.message.actualAddress
                raise ActorSystemFailure("Admin could not create requested PendingActor")
            raise ActorSystemFailure("Expected PendingActorResponse but got %s"%str(response.message))
        raise ActorSystemRequestTimeout('No response received to PendingActor request to Admin')

    def _newPrimarySendFailed(self, msg):
        self._pcrFAILED = True
        self.transport.abort_run()


    def tell(self, anActor, msg):
        self.transport.scheduleTransmit(TransmitIntent(anActor, TellMessage(msg)))  # errors ignored


    def ask(self, anActor, msg, timeout):
        self._ASKFAILED = False
        self.transport.scheduleTransmit(TransmitIntent(anActor, msg, onError = self._askSendFailed))

        if timeout is None: pass
        elif isinstance(timeout, timedelta): pass
        elif isinstance(timeout, int):
            timeout = timedelta(seconds=timeout)
        elif isinstance(timeout, float):
            timeout = timedelta(seconds=int(timeout),
                                microseconds = int((timeout - int(timeout)) * 1000 * 1000))
        else:
            raise TypeError('Unknown type for timeout: %s'%type(timeout))

        response = self.transport.run(None, timeout)
        if self._ASKFAILED:
            raise ActorSystemFailure('Transmit of ask message to %s failed (%s)'%(
                str(anActor),
                str(self._ASKFAILED)))
        if response:
            return response.message
        return None

    def _askSendFailed(self, msg):
        self._ASKFAILED = msg.result
        self.transport.abort_run()
