"Common elements for all Actors and Admin elements, regardless of implementation"

import logging
from thespian.actors import *
from thespian.system.systemBase import thesplog
from thespian.system.addressManager import ActorAddressManager, CannotPickleAddress
from thespian.system.messages import TellMessage


class systemCommonBase(object):

    def __init__(self, adminAddr, transportClass, endpoint):
        self._adminAddr   = adminAddr
        self.transport    = transportClass(endpoint)
        self._addrManager = ActorAddressManager(self.transport.myAddress)
        self._finalTransmitPending = {} # key = target ActorAddress, value=None or the last pending Intent
        self._awaitingAddressUpdate = {}  # key = actorAddress waited on (usually local), value=array of transmit Intents
        self._receiveQueue = []  # array of ReceiveMessage to be processed


    @property
    def address(self): return self.transport.myAddress
    @property
    def myAddress(self): return self.transport.myAddress


    def _handleChildExited(self, childAddress):
        self._addrManager.deadAddress(childAddress)
        self._childExited(childAddress)
        if hasattr(self, '_exiting') and not list(self._addrManager.sendingAddresses()):
            # OK, all children are dead, can now exit this actor
            self._sayGoodbye()
            self.transport.abort_run(drain=True)
        return True


    # ----------------------------------------------------------------------
    # Transmit management

    def _send_intent(self, intent):
        # Check if there are any existing transmits in progress to
        # this address (on either the input address or the validated
        # address); if so, just add the new one to the list and
        # return.
        sendAddr = self._addrManager.sendingAddress(intent.targetAddr)
        finalIntent = self._finalTransmitPending.get(intent.targetAddr,
                                                       self._finalTransmitPending.get(sendAddr, None))
        self._finalTransmitPending[sendAddr or intent.targetAddr] = intent
        if finalIntent:
            thesplog('New Intent %s to %s is waiting on previous intent %s', intent, sendAddr or intent.targetAddr, finalIntent)
            finalIntent.nextIntent = intent
            return
        self._send_intent_to_transport(intent)

    def _send_intent_to_transport(self, intent):
        thesplog('Attempting intent %s to %s', intent, intent.targetAddr)
        try:
            self.transport.scheduleTransmit(intent)
        except CannotPickleAddress as ex:
            thesplog('CannotPickleAddress, appending intent for %s (hash=%s)', ex.address, hash(ex.address))
            self._awaitingAddressUpdate.setdefault(ex.address, []).append(intent)
        except Exception:
            import traceback
            thesplog('Declaring transmit of %s to %s as Poison: %s', intent.message, intent.targetAddr, traceback.format_exc(), exc_info=True)
            if not isinstance(intent.message, TellMessage):
                self._receiveQueue.append(ReceiveEnvelope(intent.targetAddr, PoisonMessage(intent.message)))

    def _checkNextTransmit(self, completedIntent):
        if completedIntent.nextIntent:
            self._send_intent_to_transport(completedIntent.nextIntent)
        else:
            if completedIntent.targetAddr in self._finalTransmitPending:
                finalIntent = self._finalTransmitPending[completedIntent.targetAddr]
                del self._finalTransmitPending[completedIntent.targetAddr]
            else:
                sendAddr = self._addrManager.sendingAddress(completedIntent.targetAddr)
                if sendAddr in self._finalTransmitPending:
                    finalIntent = self._finalTransmitPending[sendAddr]
                    del self._finalTransmitPending[sendAddr]
                else:
                    thesplog('Completed Transmit Intent %s for unrecorded destination %s / %s',
                             str(completedIntent.message),
                             str(completedIntent.targetAddr),
                             str(sendAddr),
                             level=logging.WARNING)
                    return
            if finalIntent != completedIntent:
                thesplog('Completed final intent %s does not match recorded final intent: %s',
                         str(completedIntent), str(finalIntent),
                         level=logging.WARNING)

