"Base definitions for transports based on IP networking."


import socket


def cmpIP2Tuple(af, socktype, proto, t1, t2):
    """Function to compare two IP 2-tuple addresses.  Direct equality is
       easiest, but there are several additional equalities for the
       first element: '', '0.0.0.0', '127.0.0.1', any localIP address.
       Also, a port of 0 or None should match any other port.
    """
    if t1 == t2: return True  # easiest
    localIDs = [ None, '', '127.0.0.1' ] + \
               [ rslt[4][0]
                 for usage in [0, socket.AI_PASSIVE]
                 for useAddr in [None, socket.gethostname(), socket.getfqdn()]
                 for rslt in socket.getaddrinfo(useAddr, 0, af, socktype, proto, usage)
               ]
    if t1[0] in localIDs and t2[0] in localIDs:
        # Got a match, compare ports
        return t1[1] == t2[1] or t1[1] in [None, 0] or t2[1] in [None, 0]
    return False



class IPActorAddress(object):
    def __init__(self, af, socktype, proto, baseaddr, port, external=False):
        self.af       = af
        self.socktype = socktype
        self.proto    = proto
        if baseaddr and external and baseaddr == '0.0.0.0': baseaddr = None
        res = socket.getaddrinfo(baseaddr, port, af, socktype, proto,
                                 socket.AI_PASSIVE if baseaddr is None and not external else 0)
        af, socktype, proto, canonname, sa = res[0]
        self.sockname = sa
    def __eq__(self, o):
        return self.af == o.af and self.socktype == o.socktype and self.proto == o.proto and \
            cmpIP2Tuple(self.af, self.socktype, self.proto, self.sockname, o.sockname)
    def __ne__(self, o): return not self.__eq__(o)
    def __hash__(self): return hash(str(self))
    def __str__(self):
        if self.af == socket.AF_INET:
            if self.socktype == socket.SOCK_STREAM:
                if self.proto == socket.IPPROTO_TCP:
                    return '(TCP|%s:%d)'%self.sockname
            if self.socktype == socket.SOCK_DGRAM:
                if self.proto == socket.IPPROTO_UDP:
                    return '(UDP|%s:%d)'%self.sockname
        if self.af == socket.AF_INET6:
            if self.socktype == socket.SOCK_STREAM:
                if self.proto == socket.IPPROTO_TCP:
                    return '(TCP6|[%s]:%d %d %d)'%self.sockname
        return '(%s)'%str( ((self.af, self.socktype, self.proto), self.sockname) )
    @property
    def socketArgs(self): return (self.af, self.socktype, self.proto)
    @property
    def bindArgs(self): return self.sockname,
    @property
    def connectArgs(self): return self.sockname,


class UDPv4ActorAddress(IPActorAddress):
    def __init__(self, initialIPAddr=None, initialIPPort=0, external=False):
        super(UDPv4ActorAddress, self).__init__(socket.AF_INET,
                                                socket.SOCK_DGRAM,
                                                socket.IPPROTO_UDP,
                                                initialIPAddr, initialIPPort, external)
    def __str__(self): return '(UDP|%s:%d)'%self.sockname


class TCPv4ActorAddress(IPActorAddress):
    def __init__(self, initialIPAddr=None, initialIPPort=0, external=False):
        super(TCPv4ActorAddress, self).__init__(socket.AF_INET,
                                                socket.SOCK_STREAM,
                                                socket.IPPROTO_TCP,
                                                initialIPAddr, initialIPPort, external)
    def __str__(self): return '(TCP|%s:%d)'%self.sockname


class TCPv6ActorAddress(IPActorAddress):
    def __init__(self, initialIPAddr=None, initialIPPort=0, external=False):
        super(TCPv6ActorAddress, self).__init__(socket.AF_INET6,
                                                socket.SOCK_STREAM,
                                                socket.IPPROTO_TCP,
                                                initialIPAddr, initialIPPort, external)
    def __str__(self): return '(TCP6|[%s]:%d %d %d)'%self.sockname
