"""Uses the python multiprocess.Queue as the transport mechanism.

Queues are multi-producer/multi-consumer objects.  In this usage,
there will be only one consumer (the current Actor) although there may
be multiple producers (any other actor sending to this actor); this
actor has a single Queue for all incoming messages this actor will
handle.

However, Queues can only be passed by inheritance, not by pickling, so
only the Parent Actor has the proper queue handle to talk to the Child
Actor.  Any other Actor wishing to talk to the Child Actor will have
received the Child Actor's address from the Parent Actor, and that
address will indicate that the message must be forwarded through the
Parent Actor.  By extension, passing an Actor address itself must be
forwarded through the parent, so this means that a message passed from
one Actor to another must be passed up from the sender to it's Parent
and the message will recurse up to the Parent that is common to both
the Sender and the Receiver, where it will then traverse down the
chain of children to the destination Actor.

This also requires Parent Actors to learn about actors their Children
have created.  Each Actor therefore will maintain a table of known
addresses.  Any time an Actor creates a new child Actor, it will send
the Address of that Child Actor up to it's Parent, which will record
the new Address and the creating Child for future forwarding to that
Address.  The Parent will then recursively send the address upwards;
the Admin will end up having a table of all known Actors.

Worst case is when an actor address has been passed inside a message,
so a particular Actor has no idea how to route an address.  If this
routing deferall propagates up to the Admin, then the message will be
sent to EACH AND EVERY child (on the assumption that it will be
discarded by all leaves except the ultimate recipient).  Very
inefficient.

This transport is therefore not the most efficient transport, but it
is used as a semi-academic exercise to ensure that the Thespian system
remains flexible to handle non-socket transport mechanisms.

"""


import logging
from thespian.system.systemBase import thesplog
from thespian.actors import *
from thespian.system import timePeriodSeconds
from thespian.system.transport import *
from thespian.system.transport.asyncTransportBase import asyncTransportBase
from thespian.system.addressManager import ActorLocalAddress
from multiprocessing import Queue
import Queue as Q
from datetime import datetime
import pickle


MAX_ADMIN_QUEUESIZE=40  # depth of Admin queue
MAX_ACTOR_QUEUESIZE=10  # depth of Actor queue


class QueueActorAddress(object):  # internal use by this module only
    def __init__(self, name):
        self._qaddr = name
    def __str__(self):   return 'Q.'+str(self._qaddr)
    def __eq__(self, o): return isinstance(o, QueueActorAddress) and self._qaddr == o._qaddr
    def __ne__(self, o): return not self.__eq__(o)
    def __hash__(self): return hash(str(self))
    def subAddr(self, inst):
        import string
        useable = string.ascii_letters
        namegen = lambda v: addnext(*divmod(v, len(useable)))
        addnext = lambda x,y: useable[y] if 0 == x else (namegen(x) + useable[y])
        return self._qaddr + '.' + namegen(inst)


class MpQTEndpoint(TransportInit__Base):  # internal use by this module only
    def __init__(self, *args): self.args = args
    @property
    def addrInst(self): return self.args[0]


class MultiprocessQueueTransport(asyncTransportBase):
    """A transport designed to use a multiprocess.Queue instance to send
       and receive messages with other multiprocess Process actors.
       There is one instance of this object in each Actor.  This
       object maintains a single input queue (used by its parent an
       any children it creates) and a table of all known sub-Actor
       addresses and their queues (being the most immediate child
       Actor queue that moves the message closer to the child target,
       or the Parent actor queue to which the message should be
       forwarded if no child is identified).
    """

    def __init__(self, initType, *args):
        if isinstance(initType, ExternalInterfaceTransportInit):
            # External process that's going to talk "in".  There is no
            # parent, and the child is the systemAdmin.
            capabilities, logDefs = args
            # KWQ: do something with logDefs?
            self._parentQ         = None
            self._adminQ          = Queue(MAX_ADMIN_QUEUESIZE)
            self._adminAddr       = self.getAdminAddr(capabilities)
            self._myQAddress      = ActorAddress(QueueActorAddress('~'))
            self._myInputQ        = Queue(MAX_ACTOR_QUEUESIZE)
        elif isinstance(initType, MpQTEndpoint):
            _addrInst, myAddr, myQueue, parentQ, adminQ, adminAddr = initType.args
            self._parentQ    = parentQ
            self._adminQ     = adminQ
            self._adminAddr  = adminAddr
            self._myQAddress = myAddr
            self._myInputQ   = myQueue
        else:
            thesplog('MultiprocessQueueTransport init of type %s unsupported!', str(initType),
                     level=logging.ERROR)

        # _queues is a map of direct child ActorAddresses to Queue instance.  Note
        # that there will be multiple keys mapping to the same Queue
        # instance because routing is only either to the Parent or to
        # an immediate Child.
        self._queues = {}

        # _fwdvia represents routing for other than immediate parent
        # or child (there may be multiple target addresses mapping to
        # the same forward address.
        self._fwdvia = {} # key = targetAddress, value=fwdViaAddress

        self._nextSubInstance = 0


    @property
    def myAddress(self): return self._myQAddress


    @staticmethod
    def getAdminAddr(capabilities):
        return ActorAddress(QueueActorAddress('A')) # capabilities.get('Queue Name', 'ThespianQ')))


    def _nextSubAddress(self):
        subAddrStr = self._myQAddress.addressDetails.subAddr(self._nextSubInstance)
        self._nextSubInstance = self._nextSubInstance + 1
        return ActorAddress(QueueActorAddress(subAddrStr))


    def prepEndpoint(self, assignedLocalAddr):
        """In the parent, prepare to establish a new communications endpoint
           with a new Child Actor.  The result of this call will be
           passed to a created child process to use when initializing
           the Transport object for that class; the result of this
           call will also be kept by the parent to finalize the
           communications after creation of the Child by calling
           connectEndpoint() with this returned object.
        """
        if isinstance(assignedLocalAddr.addressDetails, ActorLocalAddress):
            return MpQTEndpoint(assignedLocalAddr.addressDetails.addressInstanceNum,
                                self._nextSubAddress(),
                                Queue(MAX_ACTOR_QUEUESIZE),
                                self._myInputQ, self._adminQ, self._adminAddr)
        return MpQTEndpoint(None,
                            assignedLocalAddr,
                            self._adminQ,         #KWQ: this passes adminQ twice because it assumes that a non-ActorLocalAddress is the admin init... that's probably not always true.
                            self._myInputQ, self._adminQ, self._adminAddr)

    def connectEndpoint(self, endPoint):
        """Called by the Parent after creating the Child to fully connect the
           endpoint to the Child for ongoing communications."""
        _addrInst, childAddr, childQueue, _myQ, _adminQ, _adminAddr = endPoint.args
        self._queues[childAddr] = childQueue


    def run(self, incomingHandler, maximumDuration=None):
        """Core scheduling method; called by the current Actor process when
           idle to await new messages (or to do background
           processing).
        """
        self._aborting_run = False
        time_to_quit = None if maximumDuration is None else datetime.now() + maximumDuration
        while (time_to_quit is None or datetime.now() < time_to_quit) and not self._aborting_run:
            try:
                rcvd = self._myInputQ.get(True,
                                          None if time_to_quit is None else
                                          timePeriodSeconds(time_to_quit -
                                                            datetime.now()))
            except Q.Empty:
                return
            relayAddr, (sendAddr, destAddr, msg) = rcvd
            if sendAddr not in self._queues:
                if relayAddr and relayAddr in self._queues:
                    self._fwdvia[sendAddr] = relayAddr
            if destAddr == self._myQAddress:
                if incomingHandler is None:
                    return ReceiveEnvelope(sendAddr, msg)
                if not incomingHandler(ReceiveEnvelope(sendAddr, msg)):
                    return  # handler returned False, indicating run() should exit
            else:
                if destAddr in self._queues:
                    self._queues[destAddr].put((self.myAddress, rcvd[1]), True)
                elif destAddr in self._fwdvia:
                    self._queues[self._fwdvia[destAddr]].put((None, rcvd[1]))
                else:
                    if relayAddr:
                        (self._parentQ or self._adminQ).put((self.myAddress, rcvd[1]), True)
                    else:
                        # Sent by parent, so this may be some
                        # grandchild not currently known.  Do the
                        # worst case and just send this message to ALL
                        # immediate children, hoping it will get there
                        # via some path.
                        for each in self._queues:
                            if each not in [self._adminAddr, str(self._adminAddr)]:
                                # None means sent by parent, so don't send BACK to parent if unknown
                                self._queues[each].put((None, rcvd[1]), True)
        return None


    def abort_run(self, drain=False):
        # Queue transmits immediately, so no draining needed
        self._aborting_run = True


    def serializer(self, intent):
        wrappedMsg = self._myQAddress, intent.targetAddr, intent.message
        # For multiprocess Queues, the serialization (pickling) of the
        # outbound message happens in a separate process.  This is
        # unfortunate because if the message is not pickle-able, the
        # exception is thrown (and not handled) in the other process,
        # and this process has no indication of the issue.  The
        # unformate solution is that pickling must be tried in the
        # current process first to detect these errors (unfortunate
        # because that means each message gets pickled twice,
        # impacting performance).
        discard = pickle.dumps(wrappedMsg)
        return wrappedMsg

    def _scheduleTransmitActual(self, transmitIntent):
        if transmitIntent.targetAddr in self._queues:
            self._queues[transmitIntent.targetAddr].put((self._myQAddress, transmitIntent.serMsg), True)
        else:
            # None means sent by parent, so don't send BACK to parent if unknown
            topOrFromBelow = self._myQAddress if self._parentQ else None
            (self._parentQ or self._adminQ).put((topOrFromBelow, transmitIntent.serMsg), True)

        transmitIntent.result = SendStatus.Sent
        transmitIntent.completionCallback()


# KWQ: test to create A->B->C and B send's C's address back inside message so A doesn't know, then send message to C and verify it gets there, with C sending back, then send to C again and verify only ONE gets there (ensuring no multipathing to C, from both A and B)
