"""Simple TCP sockets.

Each Actor has a TCP IPv4 port/socket that will accept incoming
connections for messages.  Each connection from a remote Actor will
accept a single message per connection.  The connection is dropped and
re-established for multiple messages; this is less efficient but has
more fairness.

This transport can be used within a process, between processes, and
even between processes on separate systems.

"""

# n.b. The core of this is very similar to asyncore/asynchat.
# Unfortunately, those modules are deprecated in Python 3.4 in favor
# of asyncio, which is powerful... and complex.  Thespian aims to
# support Python 2.6 through 3.4 and beyond, and has more specific
# needs (undoubtably a subset of asyncio capabilities) that can be
# implemented more simply and directly here.  In addition, this module
# should be extensible to support SSL, which asynchat is not (maybe).

# For Thespian, there are two classes of sockets:
#   * the Actor's primary receive socket, and
#   * transient outgoing send sockets.
# All steps of both types of sockets are handled asynchronously, very
# similarly to the asyncore channel.
#
# For the receive socket, it will listen for and accept incoming
# connections, and then accept a single message on that connection,
# closing the connection on completion (or error).
#
# For the transmit socket, it will connect, send, and close with a
# TransmitIntent.

# ----------------------------------------------------------------------

# TCP Buffering issues
#
# TCP is unique in that there are unusual buffering considerations to
# account for.  Specifically, a sender can connect to a listener, send
# a message, and close the socket --- *WITHOUT the receiver even
# processing the accept!  As a result, the transmitter must take
# additional steps to ensure that the message that was sent has been
# delivered.
#
# There are two ways that this confirmation can be handled:
#
#  1) confirmation sent back in the original connection
#
#  2) messages confirmed by a separate confirmation message with a
#     unique message identifier for idempotency.
#
# Disadvantages of #1:
#   * More complicated exchange between sender and receiver
#   * There must be a header synchronization with a size indicator so
#     that the receiver knows when the full message has been received
#     and should be acknowledged.
#   * The socket must exist for a potentially much longer period and
#     retransmits must still be attempted on failure.
#
# Disadvantages of #2:
#   * Doubles connection establishment requirements.
#   * More complicated send queuing to ensure ordering of messages between
#     sender and recipient.  However, this really must exist for condition
#     #1 as well.
#
# Could also do a hybrid of both.  On send, start with header
# containing message ID (and size?) then wait a brief time after send
# for the ACK, then disconnect and wait for the separate ACK later.
#
# At this point, the connection establishment seems to be the
# overriding performance dominator, and the message header
# synchronization and size indication seem like good ideas anyhow to
# confirm that the entire message has been received by the recipient.
# This method is feasible because of the asynchronous handling of the
# transmit sequence (as opposed to a blocking transmit, which would
# consume the processing budget for highly active scenarios).


import logging
from thespian.system.systemBase import thesplog
from thespian.system import timePeriodSeconds
from thespian.actors import *
from thespian.system.transport import *
from thespian.system.transport.IPBase import TCPv4ActorAddress
from thespian.system.transport.streamBuffer import (toSendBuffer, ReceiveBuffer,
                                                    ackMsg, ackPacket,
                                                    ackDataErrMsg, ackDataErrPacket)
from thespian.system.transport.asyncTransportBase import asyncTransportBase
from thespian.system.addressManager import ActorLocalAddress
import socket
import select
from datetime import datetime, timedelta
#import json
import pickle
import errno


serializer = pickle
# json cannot be used because Messages are often structures, which cannot be converted to JSON.

LISTEN_DEPTH=100  # max # of listens to sign up for at a time
MAX_INCOMING_SOCKET_PERIOD=timedelta(seconds=7)  # max time to hold open an incoming socket


class TCPEndpoint(TransportInit__Base):
    def __init__(self, *args): self.args = args
    @property
    def addrInst(self): return self.args[0]


class TCPIncoming(object):
    def __init__(self, acceptResponse):
        self._aResp = acceptResponse
        self._rData = ReceiveBuffer(serializer.loads)
        self._expires = datetime.now() + MAX_INCOMING_SOCKET_PERIOD
    @property
    def socket(self): return self._aResp[0]
    @property
    def fromAddress(self): return ActorAddress(TCPv4ActorAddress(*(self._aResp[1])))
    def delay(self):
        now = datetime.now()
        return timedelta(seconds=0) if now > self._expires else self._expires - now
    def addData(self, newData): self._rData.addMore(newData)
    def remainingSize(self): return self._rData.remainingAmount()
    def receivedAllData(self): return self._rData.isDone()
    @property
    def data(self): return self._rData.completed()[0]
    def __del__(self): self.socket.close()
    def __str__(self): return 'TCPInc(%s)<%s>'%(str(self._aResp), str(self._rData))


class TCPTransport(asyncTransportBase):
    "A transport using TCP IPv4 sockets for communications."

    def __init__(self, initType, *args):
        super(TCPTransport, self).__init__()

        if isinstance(initType, ExternalInterfaceTransportInit):
            # External process that is going to talk "in".  There is
            # no parent, and the child is the systemAdmin.
            capabilities, logDefs = args
            templateAddr          = ActorAddress(TCPv4ActorAddress(None, 0))
            self._adminAddr       = self.getAdminAddr(capabilities)
            self._parentAddr      = None
        elif isinstance(initType, TCPEndpoint):
            instanceNum, assignedAddr, self._parentAddr, self._adminAddr = initType.args
            templateAddr = assignedAddr or ActorAddress(TCPv4ActorAddress(None, 0))

        else:
            thesplog('TCPTransport init of type %s unsupported', type(initType), level=logging.ERROR)
            raise ActorSystemStartupFailure('Invalid TCPTransport init type (%s)'%type(initType))

        self.socket = socket.socket(*templateAddr.addressDetails.socketArgs)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.bind(*templateAddr.addressDetails.bindArgs)
        self.socket.listen(LISTEN_DEPTH)
        # N.B.  myAddress is actually the address we will export for
        # others to talk to us, not the bind address.  The difference
        # is that we bind to '0.0.0.0' (inaddr_any), but that's not a
        # valid address for people to send stuff to us.
        self.myAddress = ActorAddress(TCPv4ActorAddress(*self.socket.getsockname(), external=True))
        self._transmitIntents = []
        self._incomingSockets = []
        self._incomingEnvelopes = []

    def __del__(self):
        if hasattr(self, 'socket') and self.socket:
            try:
                self.socket.shutdown(socket.SHUT_RDWR)
            except: pass
            self.socket.close()


    @staticmethod
    def getAdminAddr(capabilities):
        return ActorAddress(TCPv4ActorAddress(None, capabilities.get('Admin Port', 1900)))


    def prepEndpoint(self, assignedLocalAddr):
        """In the parent, prepare to establish a new communications endpoint
           with a new Child Actor.  The result of this call will be
           passed to a created child process to use when initializing
           the Transport object for that class; the result of this
           call will also be kept by the parent to finalize the
           communications after creation of the Child by calling
           connectEndpoint() with this returned object.
        """
        if isinstance(assignedLocalAddr.addressDetails, ActorLocalAddress):
            return TCPEndpoint(assignedLocalAddr.addressDetails.addressInstanceNum,
                               None,
                               self.myAddress,
                               self._adminAddr)
        return TCPEndpoint(None,
                           assignedLocalAddr,  # assumed to be an actual TCPActorAddress-based address (e.g. admin)
                           self.myAddress,
                           self._adminAddr)

    def connectEndpoint(self, endPoint):
        pass

    _XMITStepSendConnect    = 1
    _XMITStepSendData       = 2
    _XMITStepShutdownWrite  = 3
    _XMITStepWaitForAck     = 4
    _XMITStepCloseAndFinish = 5
    _XMITStepRetry          = 6

    def serializer(self, intent):
        return toSendBuffer((self.myAddress, intent.message), serializer.dumps)

    def _scheduleTransmitActual(self, intent):
        thesplog('schedXmt.1')
        intent.stage = self._XMITStepSendConnect
        self._transmitIntents.append(intent)
        self._nextTransmitStep(intent)

    def _nextTransmitStepCheck(self, intent, fileno):
        if intent.socket.fileno() == fileno:
            return self._nextTransmitStep(intent)
        if not intent.delay():
            # Transmit timed out (consider this permanent)
            thesplog('Transmit attempt from %s to %s timed out, returning PoisonPacket',
                     self.myAddress, intent.targetAddr)
            #self._incomingEnvelopes.append(ReceiveEnvelope(intent.targetAddr,
            #                                               PoisonPacket(intent.message)))
            # Stop attempting this transmit
            intent.result = SendStatus.Failed
            intent.completionCallback()
            return False
        # Continue to attempt this transmit
        return True

    def _nextTransmitStep(self, intent):
        inputStage = intent.stage  # Select on this: only one stage processed per call
        thesplog('schedXmt.next @ %s'%inputStage)
        if self._XMITStepCloseAndFinish == inputStage:
            intent.socket.close()
            intent.completionCallback()
            return False

        if self._XMITStepSendConnect == inputStage:
            intent.socket = socket.socket(*intent.targetAddr.addressDetails.socketArgs)
            intent.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            #intent.socket.settimeout(intent.timeout)  (seconds or None)
            try:
                intent.socket.connect(*intent.targetAddr.addressDetails.connectArgs)
            except socket.error as err:
                thesplog('Socket connect failure %s', str(err))
                intent.result = SendStatus.Failed
                intent.completionCallback()
                return False
            intent.stage = self._XMITStepSendData # When connect completes

        elif self._XMITStepShutdownWrite == inputStage:
            try:
                intent.socket.shutdown(socket.SHUT_WR)
            except socket.error:
                # No shutdown handling, just close
                thesplog('schedXmt.12')
                intent.stage = self._XMITStepCloseAndFinish
                return self._nextTransmitStep(intent)
            intent.ackbuf = ReceiveBuffer(serializer.loads)
            intent.stage = self._XMITStepWaitForAck

        elif self._XMITStepSendData == inputStage:
            # Got connected, ready to send
            try:
                intent.socket.sendall(intent.serMsg)
            except Exception:
                import traceback
                thesplog('Error sending: %s', traceback.format_exc(), level=logging.ERROR)
                intent.stage = self._XMITStepRetry
                return self._nextTransmitStep(intent)
            intent.stage = self._XMITStepShutdownWrite  # After data is sent, stop transmit

        elif self._XMITStepWaitForAck == inputStage:
            # Actually, select below waited on readable, not writeable
            try:
                rcv = intent.socket.recv(intent.ackbuf.remainingAmount())
            except socket.error as err:
                if errno.ECONNRESET != err.errno:
                    raise
                thesplog('Remote %s closed connection before ack received at %s',
                         str(intent.targetAddr), str(self.myAddress),
                         level=logging.WARNING)
                rcv = ''  # Remote closed connection
            if rcv:
                intent.ackbuf.addMore(rcv)
                if not intent.ackbuf.isDone():
                    # Continue waiting for ACK
                    return True
                ackmsg = intent.ackbuf.completed()[0]
                thesplog('schedXmt.16 %s', ackmsg)
                if ackmsg in [ackPacket, ackDataErrPacket]:
                    intent.result = SendStatus.Sent if ackmsg == ackPacket else SendStatus.BadPacket
                    intent.stage = self._XMITStepCloseAndFinish
                    return self._nextTransmitStep(intent)
                thesplog('Unrecognized ACK packet')
            # Invalid ack, or no receive but socket closed.  Reschedule transmit.
            thesplog('Rescheduling transmit')
            intent.stage = self._XMITStepRetry
            return self._nextTransmitStep(intent)

        elif self._XMITStepRetry == inputStage:
            if hasattr(intent, 'socket'): delattr(intent, 'socket')
            if hasattr(intent, 'ackbuf'): delattr(intent, 'ackbuf')
            intent.stage = self._XMITStepSendConnect
            # Maybe a little time delay here?  And a retry count to give-up?
            return self._nextTransmitStep(intent)


        # Continue waiting for this transmit to complete
        return True


    def run(self, incomingHandler, maximumDuration=None):
        time_to_quit = None if maximumDuration is None else datetime.now() + maximumDuration
        while (time_to_quit is None or datetime.now() < time_to_quit) and \
              (not hasattr(self, '_aborting_run') or
               (self._aborting_run and self._transmitIntents)):
            delay = reduce(lambda d,t: t.delay() if d is None else min(d, t.delay()),
                           self._transmitIntents,
                           None if time_to_quit is None else time_to_quit - datetime.now())
            waitForSendable = lambda i, self=self: i.stage != self._XMITStepWaitForAck
            socketFile = lambda i: i.socket.fileno()
            partitionList = lambda t, f: lambda ll, i: (ll[0]+[f(i)], ll[1]) if t(i) else (ll[0], ll[1]+[f(i)])  #KWQ: standard for this?
            wsend, wrecv = reduce(partitionList(waitForSendable, socketFile),
                                  self._transmitIntents, ([], []))
            wrecv.extend([each.socket.fileno() for each in self._incomingSockets])
            delay = reduce(lambda d,i: i.delay() if d is None else min(d, i.delay()),
                           self._incomingSockets, delay)
            if not hasattr(self, '_aborting_run'):
                wrecv.extend([self.socket.fileno()])
            thesplog('select(%s, %s, [], %s) (%s %s)', wrecv, wsend, 'forever' if delay is None else timePeriodSeconds(delay), len(self._transmitIntents), getattr(self, '_aborting_run', 'no abort'))
            rrecv, rsend, _ign2 = select.select(wrecv, wsend, [],
                                                None if delay is None else timePeriodSeconds(delay))

            # Handle newly sendable data
            for eachs in rsend:
                self._transmitIntents = [I for I in self._transmitIntents
                                         if self._nextTransmitStepCheck(I, eachs)]

            # Handle newly receivable data
            for each in rrecv:
                if each == self.socket.fileno():
                    self._acceptNewIncoming()
                    continue
                self._incomingSockets = [S for S in self._incomingSockets
                                         if self._handlePossibleIncoming(S, each, incomingHandler)]
                self._transmitIntents = [I for I in self._transmitIntents
                                         if self._nextTransmitStepCheck(I, each)]

            # Handle timeouts
            self._transmitIntents = [I for I in self._transmitIntents
                                     if self._nextTransmitStepCheck(I, -1)]
            self._incomingSockets = [S for S in self._incomingSockets
                                     if self._handlePossibleIncoming(S, -1, incomingHandler)]

            # Check if it's time to quit
            if [] == rrecv and [] == rsend:
                if [] == _ign2 and time_to_quit and datetime.now() >= time_to_quit:
                    # Timeout, give up
                    thesplog('timeout run return')
                    return
                continue

            # Handle queued internal "received" data
            while self._incomingEnvelopes:
                rEnv = self._incomingEnvelopes.pop(0)
                if incomingHandler is None:
                    thesplog('TCP.run returning envelope %s', rEnv)
                    return rEnv
                if not incomingHandler(rEnv):
                    thesplog('TCP.run returning envelope %s handling returned false', rEnv)
                    return None


        thesplog('while done run return: %s, %s, %s, %s',
                 time_to_quit,
                 'None' if time_to_quit is None else datetime.now() < time_to_quit,
                 getattr(self, '_aborting_run', 'notAborting'),
                 len(self._transmitIntents))
        return None


    def _acceptNewIncoming(self):
        lsock = self.socket.accept()
        self._incomingSockets.append(TCPIncoming(lsock))
        # KWQ: Should probably timestamp incoming sockets to
        # discard them if they are idle for too long


    def _handlePossibleIncoming(self, incomingSocket, fileno, incomingHandler):
        if incomingSocket.socket.fileno() == fileno or not incomingSocket.delay():
            return self._handleReadableIncoming(incomingSocket, incomingHandler)
        return incomingSocket.delay() # Continue to wait for incoming on this socket if True


    def _handleReadableIncoming(self, inc, incomingHandler):
        rdata = inc.socket.recv(inc.remainingSize())
        if not rdata:
            # Since this point is only arrived at when select() says
            # the socket is readable, this is an indicator of a closed
            # socket.  Since previous calls didn't detect
            # receivedAllData(), this is an aborted/incomplete
            # reception.  Discard it.
            return False
        inc.addData(rdata)
        if not inc.receivedAllData():
            # Continue running and monitoring this socket
            return True
        try:
            rEnv = ReceiveEnvelope(*inc.data)
        except Exception:
            import traceback
            thesplog('OUCH!  Error deserializing received data: %s', traceback.format_exc())
            inc.socket.sendall(ackDataErrMsg)
            # Continue running, but release this socket
            return False
        inc.socket.sendall(ackMsg)
        self._incomingEnvelopes.append(rEnv)
        # Continue to run, but this socket is releasable
        return False


    def abort_run(self, drain=False):
        self._aborting_run = drain
