"This module provides various low-level inter-Actor transport implementations."

from datetime import datetime, timedelta


DEFAULT_MAX_TRANSMIT_PERIOD = timedelta(minutes=5)


class TransportInit__Base(object): pass
class ExternalInterfaceTransportInit(TransportInit__Base):
    """Used as first argument to Transport __init__ to indicate that this
       is an external process interfacing to the ActorSystem via the
       Transport.
    """
    pass


# ----------------------------------------------------------------------

class ReceiveEnvelope(object):
    "Represents the message received along with the sender's address"
    def __init__(self, sender, msg):
        self._sender  = sender
        self._message = msg
    @property
    def sender(self): return self._sender
    @property
    def message(self): return self._message



# ----------------------------------------------------------------------


class TransmitIntent(object):
    """An individual transmission of data can be encapsulated by a
       "transmit intent", which identifies the message and the target
       address, and which has a callback for eventual success or
       failure indication.  Transmit intents may be chained together
       to represent a series of outbound transmits.  Adding a transmit
       intent to the chain may block when the chain reaches an upper
       threshold, and remain blocked until enough transmits have
       occured (successful or failed) to reduce the size of the chain
       below a minimum threshold.  This acts to implement server-side
       flow control in the system as a whole (although it can
       introduce a deadlock scenario if multiple actors form a
       transmit loop that is blocked at any point in the loop, so a
       transmit intent will fail if it reaches a maximum number of
       retries without success).

       The TransmitIntent is constructed with a target address, the
       message to send, and optional onSuccess and onError callbacks
       (both defaulting to None).  The callbacks are passed the
       TransmitIntent when the transport is finished with it,
       selecting the appropriate callback based on the completion
       status (the `result' property will reveal the SendStatus actual
       result of the attempt).  A callback of None will simply discard
       the TransmitIntent without passing it to a callback.

       The TransmitIntent is passed to the transport that should
       perform the intent; the transport may attach its own additional
       data to the intent during that processing.

    """

    def __init__(self, targetAddr, msg, onSuccess=None, onError=None, maxPeriod=None):
        #KWQ maxRetries?
        self._targetAddr = targetAddr
        self._message    = msg
        self._successTo  = onSuccess
        self._failureTo  = onError
        self._resultsts  = None
        self._quitTime   = datetime.now() + (maxPeriod or DEFAULT_MAX_TRANSMIT_PERIOD)
        self.nextIntent  = None

    @property
    def targetAddr(self): return self._targetAddr

    def changeTargetAddr(self, newAddr): self._targetAddr = newAddr

    @property
    def message(self): return self._message

    @property
    def result(self): return self._resultsts
    @result.setter
    def result(self, setResult):
        if not isinstance(setResult, SendStatus.BASE):
            raise TypeError('TransmitIntent result must be a SendStatus (got %s)'%type(setResult))
        self._resultsts = setResult

    def completionCallback(self):
        "This is called by the transport to perform the success or failure callback operation."
        ((self._successTo if self.result else self._failureTo) or (lambda m: None))(self)

    def delay(self):
        now = datetime.now()
        return max(timedelta(seconds=0), self._quitTime - now)


# KWQ: transmit intent should have some way to complete run if run incomingHandler is None (systemBase)  This is done by the abort_run() call.

class SendStatus(object):
    class BASE(object): pass
    class SENDSTS_SENT(BASE): pass
    class SENDSTS_NOTSENT(BASE): pass  # has not been sent, has not been actively rejected
    class BadPacketError(BASE, Exception): pass  # remote rejected, (a return value or an exception)
    class SENDSTS_FAILED(BASE): pass
    Sent = SENDSTS_SENT()
    NotSent = SENDSTS_NOTSENT()
    BadPacket = BadPacketError('BadPacket SendStatus')
    Failed = SENDSTS_FAILED()

