"""This module provides a functional association between the generic
ActorAddress and the socket-based implementation of those
ActorAddresses."""


import sys, socket, pickle, errno, time, select
from thespian.actors import *
import logging

SEND_CONFIRM = True


ADDR_INUSE_RETRIES              = 5    # of retries to bind to addr if previous user exiting
ADDR_INUSE_BACKOFF_PERIOD       = 0.13 # cumulative # of seconds to backoff before each retry
SEND_CONNREFUSED_RETRIES        = 3    # of retries if target socket in transition.
SEND_CONNREFUSED_BACKOFF_PERIOD = 0.23 # cumulative # of seconds to backoff each retry
SEND_OR_FAIL_MAX_DELAY          = 60   # max # of seconds for transmit backoff periods
ASK_SEND_TIMEOUT = 0.25
HANDSHAKE_TIMEOUT = 0.067

class SocketActorAddress:
    def __init__(self, af, socktype, proto, sockname):
        self.af       = af
        self.socktype = socktype
        self.proto    = proto
        self.sockname = sockname
    def __eq__(self, o):
        return self.af == o.af and self.socktype == o.socktype and self.proto == o.proto and \
            self.sockname == o.sockname
    def __ne__(self, o): return not self.__eq__(o)
    def __hash__(self): return hash('%s:%s:%s:%s'%(self.af, self.socktype, self.proto, self.sockname))
    def __str__(self):
        if self.af == socket.AF_INET:
            if self.socktype == socket.SOCK_STREAM:
                if self.proto == socket.IPPROTO_TCP:
                    return '(TCP|%s:%d)'%self.sockname
        if self.af == socket.AF_INET6:
            if self.socktype == socket.SOCK_STREAM:
                if self.proto == socket.IPPROTO_TCP:
                    return '(TCP6|%s:%d)'%self.sockname
        return '(%s)'%str( ((self.af, self.socktype, self.proto), self.sockname) )
    @property
    def socketArgs(self): return (self.af, self.socktype, self.proto)
    @property
    def bindArgs(self): return self.sockname,
    @property
    def connectArgs(self): return self.sockname,


def newActorAddress(port=0, name='Actor', sysaddress=None):
    #nas = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    for useAddr in [None, sysaddress or socket.gethostname(), None]:
        res = socket.getaddrinfo(useAddr, port,
                                 socket.AF_UNSPEC, socket.SOCK_STREAM, 0, socket.AI_PASSIVE)
        for af, socktype, proto, canonname, sa in res:
            try:
                # all network interfaces, system picks port
                generaladdr = ActorAddress(SocketActorAddress(af, socktype, proto, sa))
                # This isn't the process to handle connections, but
                # opening and listening now makes the
                # socket "live" until the process to handle
                # connections starts below.
                nas = addressListener(generaladdr, name)
                # actual address bound to
                addr = ActorAddress(SocketActorAddress(af, socktype, proto, nas.socket.getsockname()))
                if port != 0 and addr.addressDetails.bindArgs[0][1] != port:
                    nas.close()
                    raise InvalidActorAddress(generaladdr,
                                              'not mapped to requested port %s'%port)
                return addr,nas
            except socket.error as ex:
                pass # try the next possible address
    if res:
        invaddr = ActorAddress( "Unable to bind with %s"%res )
    else:
        invaddr = ActorAddress( "Unable to get address info" )
    raise InvalidActorAddress(invaddr,
                              'Could not generate a new actor address' + \
                              ' (system resource exhaustion?)')


class SimpleSocketWrapper(object):
    def __init__(self, baseSocket, name='socket'):
        self.socket = baseSocket
    def __enter__(self): return self
    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
        return False
    def bind(self, toAddr):
        self.socket.bind(toAddr)
        self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
    def connect(self, toAddr):
        self.socket.connect(toAddr)
        self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
    def fileno(self): return self.socket.fileno()  # for select.select
    def sendall(self, *args):
        return self.socket.sendall(*args)
    def recv(self, *args):
        return self.socket.recv(*args)
    def close(self):
        return self.socket.close()
    def accept(self, *args, **kw):
        asock, frmaddr = self.socket.accept(*args, **kw)
        rasock = SocketWrapper(asock, 'accepted')
        return rasock

class LoggingSocketWrapper(SimpleSocketWrapper):
    def __init__(self, baseSocket, name='socket'):
        super(LoggingSocketWrapper, self).__init__(baseSocket, name)
        self._rawname = name
        self._name = self._namestr()
        self.logAction('creating')
    def _namestr(self):
        return '--' + ('%s:%05d'%self.socket.getsockname()).replace('0.0.0.0:','....any..:') + '--' + self._rawname
    def logAction(self, action):
        with open('socket.log', 'a') as newfd:
            newfd.write('socket%s: %s\n'%(self._name, action))
    def bind(self, toAddr):
        self.logAction('bind attempt to %s'%str(toAddr))
        self.socket.bind(toAddr)
        self._name = self._namestr()
        self.logAction('bind')
    def connect(self, toAddr):
        self.socket.connect(toAddr)
        self._name = self._namestr()
        self.logAction('connect')
    def sendall(self, *args):
        self.logAction('sending')
        return self.socket.sendall(*args)
    def recv(self, *args):
        self.logAction('receiving')
        return self.socket.recv(*args)
    def close(self):
        self.logAction('closing')
        return self.socket.close()
    def accept(self, *args, **kw):
        asock, frmaddr = self.socket.accept(*args, **kw)
        rasock = SocketWrapper(asock, self._rawname + '-accepted')
        self.logAction('accept as %s from %s'%(rasock._name, frmaddr))
        return rasock

SocketWrapper = SimpleSocketWrapper


def addressListener(listenActorAddress, name='socket'):
    aSocket = SocketWrapper(socket.socket(*listenActorAddress.addressDetails.socketArgs), name)
    try:
        aSocket.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        for retries in range(ADDR_INUSE_RETRIES):
            try:
                aSocket.bind(*listenActorAddress.addressDetails.bindArgs)
                break
            except socket.error as ex:
                if ex.errno == errno.EADDRINUSE:  # probably transitional
                    time.sleep((retries+1) * ADDR_INUSE_BACKOFF_PERIOD)
                    continue
                else:
                    # some other bind error
                    raise
        aSocket.socket.listen(5)
    except:
        # n.b. explicit close here instead of "with" context above
        # because if no error this should return with a still active
        # socket.
        aSocket.close()
        raise
    return aSocket


from thespian.system.transport import SendStatus

# sendToAddress returns one of SendStatus values
# sendOrFail has no return value but raises BadPacket or SendTimeoutError()
# askToAddress returns response, None, or raises BadPacket


def sendToAddress(rmtAddress, msg, senderAddress, timeout=None):
    outMsg = (senderAddress,msg)
    # Allow a limited number of retries for target sockets that are in-transition.
    for retries in range(1):  #SEND_CONNREFUSED_RETRIES):
        try:
            with SocketWrapper(socket.socket(*rmtAddress.addressDetails.socketArgs),
                               'sender from %s'%str(senderAddress)) as s:
                s.socket.settimeout(timeout)
                s.connect(*rmtAddress.addressDetails.connectArgs)
                if SEND_CONFIRM:
                    # Exchange an initial handshake before sending the
                    # real message.  This is needed because a socket
                    # send will succeed even if the recipient hasn't
                    # accepted the connection.
                    try:
                        s.socket.settimeout(HANDSHAKE_TIMEOUT)
                        s.sendall('Q'.encode('utf-8'))
                        r = s.recv(1)
                        if r.decode('utf-8') != '+':
                            return False
                    except socket.error as err:
                        s.sendall('!'.encode('utf-8'))
                        return SendStatus.NotSent
                s.socket.settimeout(timeout)
                s.sendall(toSendBuffer(outMsg))
                if SEND_CONFIRM:
                    rBuf = ReceiveBuffer()
                    rCnt = 0
                    try:
                        while True:
                            rCnt = rCnt + 1
                            if rCnt > 7:
                                raise ValueError('too many acknowlegement retries')
                            try:
                                rcv = s.recv(rBuf.remainingAmount())
                                #KWQ: except socket.timeout?
                            except socket.error as err:
                                if err.errno == errno.ECONNRESET: #54
                                    rcv = ''
                                else:
                                    raise
                            if rcv == '':
                                # socket was closed, send wasn't confirmed.
                                break
                            rBuf.addMore(rcv)
                            if rBuf.isDone():
                                rbc = rBuf.completed()[0]
                                if rbc == ackPacket:
                                    return SendStatus.Sent
                                if rbc == ackDataErrPacket:
                                    return SendStatus.BadPacket
                                break  # sorry, didn't get a valid message
                    except Exception as ex:
                        raise
                else:
                   return SendStatus.Sent
                # No return, next retries loop attempt needed
        except socket.timeout:
            # Possible race condition here!  This timeout might have
            # occurred on the local side just after the remote side
            # confirmed everything, so the remote side thinks
            # everything is good... but this side doesn't.
            return SendStatus.NotSent
        except socket.error as err:
            if err.errno in [errno.ECONNREFUSED, errno.EPIPE, errno.ECONNRESET]:
                # Can happen if target was just started and hasn't
                # fully started up yet, or if target just exited and
                # deadletter handling is not yet setup.  Allow target
                # to stabilize
                #time.sleep(SEND_CONNREFUSED_BACKOFF_PERIOD * (retries+1))
                continue   # retry
            raise
        except Exception as err:
            raise
    return SendStatus.NotSent


class SendTimeoutError(Exception): pass


def sendOrFail(rmtAddress, msg, senderAddress, timeout=300):
    for delay in timeseq(timeout):
        sts = sendToAddress(rmtAddress, msg, senderAddress, 1.0)
        if sts == SendStatus.Sent: return
        if sts == SendStatus.BadPacket: raise sts
        time.sleep(delay)
    raise SendTimeoutError()


from streamBuffer import (ReceiveBuffer, toSendBuffer, ackPacket, ackDataErrPacket,
                          ackMsg, ackDataErrMsg)


def acceptAndReceive(listenSock, timeout=10, rspAddr='unk'):
    try:
        listenSock.socket.settimeout(timeout or 10)
        while True:  # retry to clean out handshake failures
            with listenSock.accept() as rSock:
                if SEND_CONFIRM:
                    try:
                        rSock.socket.settimeout(HANDSHAKE_TIMEOUT)
                        h = rSock.recv(3)
                        if h.decode('utf-8') != 'Q':
                            continue
                        rSock.sendall('+'.encode('utf-8'))
                    except socket.error as err:
                        continue
                return _Receive(rSock, timeout, rspAddr)
    except socket.timeout:
        # logger.warning('timeout internally asking %s on accept'%(str(rmtAddress)))
        pass
    return None


def _Receive(rSock, timeout, rspAddr='unk'):
    rSock.socket.settimeout(timeout or 10)
    try:
        rBuf = ReceiveBuffer()
        retry = 0
        while True:
            retry += 1
            try:
                rcv = rSock.recv(rBuf.remainingAmount())
            except socket.error as err:
                if err.errno == errno.EWOULDBLOCK:
                    time.sleep(0.1)
                    continue
                return None
            if rcv == '':
                # can still return buffer if done?
                return None  # socket was closed
            rBuf.addMore(rcv)
            if rBuf.isDone():
                try:
                    rsp = rBuf.completed()
                except Exception:
                    # Probably ImportError or PickleError.  This is an
                    # invalid message that was successfully received,
                    # so tell the remote we got it but can't use it.
                    if SEND_CONFIRM:
                        try:
                            rSock.sendall(ackDataErrMsg)
                        except Exception as ex:
                            pass # ignore errors sending ack
                    return None  # Didn't receive anything actionable

                if rsp is not None:
                    if SEND_CONFIRM:
                        try:
                            rSock.sendall(ackMsg)
                        except Exception as ex:
                            pass # ignore errors sending ack
                    return rsp[0]  # discard extra
                return None    # sorry, didn't get a valid message
    except socket.timeout:
        # logger.warning('timeout asking %s on recv'%(str(rmtAddress)))
        return None
    except Exception as ex:
        # Probably a pickling error on rbuf
        # pretty much ignore all errors receiving and pickling
        return None


def timeseq(maxTime, incr=1.3, origDelay=0.01):
    curTime = min(origDelay, SEND_OR_FAIL_MAX_DELAY)
    while maxTime is None or maxTime > 0:
        yield curTime
        if maxTime:
            maxTime = maxTime - curTime
        curTime = min(curTime * incr, SEND_OR_FAIL_MAX_DELAY)


def askToAddress(rmtAddress, msg, timeout=10):
    rspAddr, rspSock = newActorAddress(name='asker')
    logging.getLogger('Thespian').debug('asker @ %s'%str(rspAddr))
    sentOK = False
    if rspSock:
        with rspSock as responseListenerSock:
            responseListenerSock.socket.settimeout(timeout if timeout else 5)

            for delay in timeseq(timeout):
                if not sentOK:
                    sendsts = sendToAddress(rmtAddress, msg, rspAddr, ASK_SEND_TIMEOUT)
                    if sendsts == SendStatus.Sent:
                        sentOK = True
                    if sendsts == SendStatus.BadPacket:
                        raise sendsts
                    # n.b. even if sendToAddress returns false, it's
                    # possible that the remote did successfully get the
                    # message (see note re timeout in send above).
                    # Therefore, use select on response with timeout
                    # irregardless of sendToAddress result.
                sresp,ign1,ign2 = select.select([responseListenerSock.socket], [], [],
                                                timeout if sentOK else delay)
                for each in sresp:
                    resp = acceptAndReceive(responseListenerSock, timeout)
                    if resp is not None:
                        return resp[1]  # return only message, not sender address
                if sentOK: break
    return None
